# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import unittest

import processscheduler as ps


class TestCost(unittest.TestCase):
    def test_cost_basic(self) -> None:
        ps.SchedulingProblem("CostBasic", horizon=12)
        ress_cost = ps.ConstantCostPerPeriod(5)
        ps.Worker("Worker1", cost=ress_cost)

    def test_cost_polynomial_1(self) -> None:
        ps.SchedulingProblem("PolynomialCost1", horizon=12)

        def c(t):
            return 0.5 * t ** 2 + 50

        ress_cost = ps.PolynomialCostFunction(c)
        ps.Worker("Worker1", cost=ress_cost)

    def test_cost_polynomial_lambda(self) -> None:
        ps.SchedulingProblem("PolynomialCostLambdaFunction", horizon=12)
        ress_cost = ps.PolynomialCostFunction(lambda t: 2 * t + 10)
        ps.Worker("Worker1", cost=ress_cost)

    def test_cost_failure(self) -> None:
        ps.SchedulingProblem("CostWrongType", horizon=12)
        with self.assertRaises(TypeError):
            ps.Worker("ress1", cost=5)  # only accepts a _Cost instance
        with self.assertRaises(TypeError):
            ps.PolynomialCostFunction("f")  # only accepts a callable

    def test_constant_cost_per_period_1(self) -> None:
        problem = ps.SchedulingProblem("IndicatorResourceConstantCostPerPeriod1")
        t_1 = ps.FixedDurationTask("t1", duration=11)
        worker_1 = ps.Worker("Worker1", cost=ps.ConstantCostPerPeriod(7))
        t_1.add_required_resource(worker_1)
        cost_ind = problem.add_indicator_resource_cost([worker_1])

        solution = ps.SchedulingSolver(problem).solve()

        self.assertTrue(solution)
        self.assertEqual(solution.indicators[cost_ind.name], 77)

    def test_constant_cost_per_period_2(self) -> None:
        problem = ps.SchedulingProblem("IndicatorResourceConstantCostPerPeriod12")
        t_1 = ps.VariableDurationTask("t1", work_amount=100)
        worker_1 = ps.Worker(
            "Worker1", productivity=4, cost=ps.ConstantCostPerPeriod(10)
        )
        worker_2 = ps.Worker(
            "Worker2", productivity=7, cost=ps.ConstantCostPerPeriod(20)
        )
        all_workers = [worker_1, worker_2]
        problem.add_objective_makespan()
        t_1.add_required_resources(all_workers)
        cost_ind = problem.add_indicator_resource_cost(all_workers)

        solution = ps.SchedulingSolver(problem).solve()

        self.assertTrue(solution)
        self.assertEqual(solution.indicators[cost_ind.name], 300)

    def test_linear_cost_1(self) -> None:
        problem = ps.SchedulingProblem("IndicatorResourcePolynomialCost1")

        t_1 = ps.FixedDurationTask("t1", duration=17)

        def int_cost_function(t):
            return 23 * t + 3

        worker_1 = ps.Worker(
            "Worker1", cost=ps.PolynomialCostFunction(int_cost_function)
        )
        t_1.add_required_resource(worker_1)

        problem.add_constraint(ps.TaskStartAt(t_1, 13))
        cost_ind = problem.add_indicator_resource_cost([worker_1])

        solution = ps.SchedulingSolver(problem).solve()

        self.assertTrue(solution)
        # expected cost should be 8457
        expected_cost = int(
            ((int_cost_function(13) + int_cost_function(13 + 17)) * 17) / 2
        )
        self.assertEqual(solution.indicators[cost_ind.name], expected_cost)

    def test_optimize_linear_cost_1(self) -> None:
        # same cost function as above, we minimize the cst,
        # the task should be scheduled at 0 (because the cost function increases)
        problem = ps.SchedulingProblem("OptimizeLinearCost1")

        t_1 = ps.FixedDurationTask("t1", duration=17)

        def int_cost_function(t):
            return 23 * t + 3

        worker_1 = ps.Worker(
            "Worker1", cost=ps.PolynomialCostFunction(int_cost_function)
        )
        t_1.add_required_resource(worker_1)

        cost_ind = problem.add_indicator_resource_cost([worker_1])
        problem.add_objective_resource_cost([worker_1])

        solution = ps.SchedulingSolver(problem).solve()

        self.assertTrue(solution)
        # the task is scheduled at the beginning of the workplan
        self.assertEqual(solution.tasks[t_1.name].start, 0)
        # expected cost should be 3374
        expected_cost = int(((int_cost_function(0) + int_cost_function(17)) * 17) / 2)
        self.assertEqual(solution.indicators[cost_ind.name], expected_cost)

    def test_optimize_linear_cost_2(self) -> None:
        # now we have a cost function that decreases over time,
        # then minimizing the cost should schedule the task at the end.
        # so we define an horizon for the problem
        problem = ps.SchedulingProblem("OptimizeLinear2", horizon=87)

        t_1 = ps.FixedDurationTask("t1", duration=13)

        # be sure however that this function is positive on the interval
        def int_cost_function(t):
            return 711 - 5 * t

        worker_1 = ps.Worker(
            "Worker1", cost=ps.PolynomialCostFunction(int_cost_function)
        )
        t_1.add_required_resource(worker_1)

        cost_ind = problem.add_indicator_resource_cost([worker_1])
        problem.add_objective_resource_cost([worker_1])

        solution = ps.SchedulingSolver(problem).solve()

        self.assertTrue(solution)
        # the task is scheduled at the beginning of the workplan
        self.assertEqual(solution.tasks[t_1.name].start, 74)

        # expected cost should be 3374
        expected_cost = int(((int_cost_function(74) + int_cost_function(87)) * 13) / 2)
        self.assertEqual(solution.indicators[cost_ind.name], expected_cost)

    def test_optimize_linear_cost_3(self) -> None:
        # same cost function as above, we minimize the cst,
        # the task should be scheduled at 0 (because the cost function increases)
        problem = ps.SchedulingProblem("OptimizeLinearCost3")

        t_1 = ps.FixedDurationTask("t1", duration=17)

        def real_cost_function(t):
            return 23.112 * t + 3.5

        worker_1 = ps.Worker(
            "Worker1", cost=ps.PolynomialCostFunction(real_cost_function)
        )
        t_1.add_required_resource(worker_1)

        cost_ind = problem.add_indicator_resource_cost([worker_1])
        problem.add_objective_resource_cost([worker_1])

        solution = ps.SchedulingSolver(problem).solve()

        self.assertTrue(solution)
        # the task is scheduled at the beginning of the workplan
        self.assertEqual(solution.tasks[t_1.name].start, 0)
        # expected cost should be 3374
        expected_cost = ((real_cost_function(0) + real_cost_function(17)) * 17) / 2
        # Because there are Int type conversions, the result may not be exactly
        # the one expected, let's assume 99% is a good approx.
        err = abs(solution.indicators[cost_ind.name] / expected_cost - 1)
        self.assertLessEqual(err, 0.01)

    def test_quadratic_cost_1(self) -> None:
        problem = ps.SchedulingProblem("IndicatorResourceQuadraticCost1")

        t_1 = ps.FixedDurationTask("t1", duration=17)

        def int_cost_function(t):
            return 23 * t ** 2 - 13 * t + 513

        worker_1 = ps.Worker(
            "Worker1", cost=ps.PolynomialCostFunction(int_cost_function)
        )
        t_1.add_required_resource(worker_1)

        problem.add_constraint(ps.TaskStartAt(t_1, 13))
        cost_ind = problem.add_indicator_resource_cost([worker_1])

        solution = ps.SchedulingSolver(problem).solve()

        self.assertTrue(solution)
        # expected cost should be 8457
        expected_cost = int(
            ((int_cost_function(13) + int_cost_function(13 + 17)) * 17) / 2
        )
        self.assertEqual(solution.indicators[cost_ind.name], expected_cost)

    def test_optimize_quadratic_cost_2(self) -> None:
        # TODO: add an horizon, it should return the expected result
        # but there's an issue, see
        # https://github.com/Z3Prover/z3/issues/5254
        problem = ps.SchedulingProblem("OptimizeQuadraticCost2", horizon=20)

        t_1 = ps.FixedDurationTask("t1", duration=4)

        # we chosse a function where we know the minimum is
        # let's imagine the minimum is at t=37
        def int_cost_function(t):
            return (t - 8) ** 2 + 100

        worker_1 = ps.Worker(
            "Worker1", cost=ps.PolynomialCostFunction(int_cost_function)
        )
        t_1.add_required_resource(worker_1)

        cost_ind = problem.add_indicator_resource_cost([worker_1])
        problem.minimize_indicator(cost_ind)

        solution = ps.SchedulingSolver(problem).solve()

        self.assertTrue(solution)
        # TODO: to be fixed; Non linear solver gives weirf results.
        # This was repored to z3 team see
        # https://github.com/Z3Prover/z3/issues/5254
        # Fix to be released in the next z3 version 4.8.11.0
        # sometimes it's 6, sometime it's 7 TODO weird
        # self.assertTrue(solution.tasks[t_1.name].start in [6, 7])

        # sometimes it's 416, sometimes 420 TODO
        # self.assertTrue(solution.indicators[cost_ind.name] in [416, 420])

    def test_plot_cost_function(self) -> None:
        # TODO: add an horizon, it should return the expected result
        # but there's an issue, see
        # https://github.com/Z3Prover/z3/issues/5254

        # we chosse a function where we know the minimum is
        # let's imagine the minimum is at t=37
        def int_cost_function(t):
            return (t - 37) ** 2 + 513

        cost = ps.PolynomialCostFunction(int_cost_function)

        cost.plot([0, 40], show_plot=False)

    def test_cumulative_cost(self):
        problem = ps.SchedulingProblem("CumulativeCost", horizon=5)
        t_1 = ps.FixedDurationTask("t1", duration=5)
        t_2 = ps.FixedDurationTask("t2", duration=5)
        t_3 = ps.FixedDurationTask("t3", duration=5)
        worker_1 = ps.CumulativeWorker(
            "CumulWorker", size=3, cost=ps.ConstantCostPerPeriod(5)
        )
        t_1.add_required_resource(worker_1)
        t_2.add_required_resource(worker_1)
        t_3.add_required_resource(worker_1)

        cost_ind = problem.add_indicator_resource_cost([worker_1])

        solution = ps.SchedulingSolver(problem).solve()

        self.assertTrue(solution)
        self.assertEqual(solution.indicators[cost_ind.name], 25)
        solution = ps.SchedulingSolver(problem).solve()
        self.assertTrue(solution)

    def test_incremental_optimizer_linear_cost_1(self) -> None:
        # same cost function as above, we minimize the cst,
        # the task should be scheduled at 0 (because the cost function increases)
        problem = ps.SchedulingProblem("IncrementalOptimizerLinearCost1", horizon=40)

        t_1 = ps.FixedDurationTask("t1", duration=17)

        # the task should be scheduled at the end
        def int_cost_function(t):
            return -23 * t + 321

        worker_1 = ps.Worker(
            "Worker1", cost=ps.PolynomialCostFunction(int_cost_function)
        )
        t_1.add_required_resource(worker_1)

        cost_ind = problem.add_indicator_resource_cost([worker_1])
        problem.minimize_indicator(cost_ind)

        solver = ps.SchedulingSolver(problem, parallel=True, random_seed=True)

        solution = solver.solve()

        self.assertTrue(solution)
        self.assertEqual(solution.tasks[t_1.name].start, 23)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

from processscheduler.problem import SchedulingProblem
from processscheduler.resource import Worker, CumulativeWorker, SelectWorkers
from processscheduler.task import (
    ZeroDurationTask,
    FixedDurationTask,
    VariableDurationTask,
)
from processscheduler.task_constraint import (
    TaskPrecedence,
    TasksStartSynced,
    TasksEndSynced,
    TasksDontOverlap,
    TaskStartAt,
    TaskStartAfterStrict,
    TaskStartAfterLax,
    TaskEndAt,
    TaskEndBeforeStrict,
    TaskEndBeforeLax,
    OptionalTaskConditionSchedule,
    OptionalTasksDependency,
    ForceScheduleNOptionalTasks,
    ScheduleNTasksInTimeIntervals,
)
from processscheduler.resource_constraint import (
    WorkLoad,
    ResourceUnavailable,
    DistinctWorkers,
    SameWorkers,
)
from processscheduler.cost import ConstantCostPerPeriod
from processscheduler.solver import SchedulingSolver

import isodate
import ipywidgets as widgets

#
# Problem UI
#
problem_name_widget = widgets.Text(
    value="MySchedulingProblem",
    placeholder="Enter scheduling problem name",
    description="Name:",
    disabled=False,
)
set_horizon_widget = widgets.Checkbox(
    value=False,
    description="Set horizon",
    disabled=False,
    indent=False,
    layout={"width": "200px"},
)
horizon_widget = widgets.IntText(
    description="Horizon:", disabled=True, layout={"width": "200px"}
)


def on_set_horizon_clicked(b):
    horizon_widget.disabled = not set_horizon_widget.value


set_horizon_widget.observe(on_set_horizon_clicked)

create_problem_button = widgets.Button(
    description="Create problem",
    disabled=False,
    button_style="",
    tooltip="Click to create a new SchedulingProblem",
)
start_time_widget = widgets.Text(
    value="",
    placeholder="Optional, you can leave blank",
    description="Start Time:",
    disabled=False,
)
delta_time_widget = widgets.Text(
    value="",
    placeholder="Optional, you can leave blank",
    description="Delta time:",
    disabled=False,
)
problem_output = widgets.Output()


def on_create_problem_button_clicked(b) -> bool:
    global pb
    problem_name = problem_name_widget.value
    # start time
    parsed_start_time = None
    if start_time_widget.value != "":
        parsed_start_time = isodate.parse_datetime(start_time_widget.value)
        try:
            parsed_start_time = isodate.parse_datetime(start_time_widget.value)
        except ValueError:
            print("Not a valid isodate format for start_time")
            parsed_start_time = None
            return False
    # timedelta
    parsed_delta_time = None
    if delta_time_widget.value != "":
        try:
            parsed_delta_time = isodate.parse_duration(delta_time_widget.value)
        except isodate.isoerror.ISO8601Error:
            print("Not a valid isodate format for delta_time")
            parsed_delta_time = None
            return False
    if set_horizon_widget.value:
        pb = SchedulingProblem(
            problem_name,
            horizon_widget.value,
            start_time=parsed_start_time,
            delta_time=parsed_delta_time,
        )
    else:
        pb = SchedulingProblem(
            problem_name, start_time=parsed_start_time, delta_time=parsed_delta_time
        )
    with problem_output:
        print("Scheduling problem", problem_name, "successfully created.")
    create_problem_button.disabled = True
    return True


create_problem_button.on_click(on_create_problem_button_clicked)
problem_ui = widgets.VBox(
    [
        widgets.HBox(
            [
                problem_name_widget,
                widgets.VBox(
                    [
                        widgets.HBox(
                            [
                                horizon_widget,
                                set_horizon_widget,
                            ]
                        ),
                        start_time_widget,
                        delta_time_widget,
                    ],
                    layout=widgets.Layout(border="solid 1px gray"),
                ),
                create_problem_button,
            ]
        ),
        problem_output,
    ]
)
#
# Resource UI
#
resource_types = {"Worker": Worker, "CumulativeWorker": CumulativeWorker}
resource_type_widget = widgets.Dropdown(
    options=list(resource_types),
    value="Worker",
    description="Resource type:",
    disabled=False,
)
resource_name_widget = widgets.Text(
    value="",
    placeholder="Enter resource name",
    description="Name:",
    disabled=False,
    layout={"width": "200px"},
)
resource_cost_per_period_widget = widgets.IntText(
    value=1, disabled=False, description="Cost/Period:", layout={"width": "200px"}
)
resource_productivity_widget = widgets.IntText(
    value=0,
    disabled=False,
    description="Productivity/Period:",
    layout={"width": "200px"},
)

resource_size_widget = widgets.IntText(
    value=2, disabled=True, description="Size:", layout={"width": "200px"}
)


def on_change_resource_type(change):
    # print("popo", change)
    if change["type"] == "change" and change["name"] == "value":
        new_value = "%s" % change["new"]
        if new_value == "CumulativeWorker":
            resource_size_widget.disabled = False
        elif new_value == "Worker":
            resource_size_widget.disabled = True


resource_type_widget.observe(on_change_resource_type)

create_resource_button = widgets.Button(
    description="Create resource",
    disabled=False,
    button_style="",
    tooltip="Click to create a new Resource",
)
resource_output = widgets.Output()


def on_create_resource_button_clicked(b):
    resource_name = resource_name_widget.value
    resource_type = resource_types[resource_type_widget.value]
    if resource_type == Worker:
        new_resource = resource_type(
            resource_name,
            productivity=resource_productivity_widget.value,
            cost=ConstantCostPerPeriod(resource_cost_per_period_widget.value),
        )
    elif resource_type == CumulativeWorker:
        new_resource = resource_type(
            resource_name,
            size=resource_size_widget.value,
            productivity=resource_productivity_widget.value,
            cost=ConstantCostPerPeriod(resource_cost_per_period_widget.value),
        )
    # rebuild option list for the task list
    resources_list_of_tuples = []
    for resource in pb.context.resources:
        resources_list_of_tuples.append((resource.name, resource))
    resources_select_widget.options = resources_list_of_tuples
    with resource_output:
        print("Resource", new_resource.name, "successfully created.")


create_resource_button.on_click(on_create_resource_button_clicked)
resource_ui = widgets.VBox(
    [
        widgets.HBox(
            [
                resource_type_widget,
                widgets.VBox(
                    [
                        resource_name_widget,
                        resource_cost_per_period_widget,
                        resource_productivity_widget,
                        resource_size_widget,
                    ]
                ),
                create_resource_button,
            ]
        ),
        resource_output,
    ]
)

#
# Task UI
#
task_types = {
    "ZeroDurationTask": ZeroDurationTask,
    "FixedDurationTask": FixedDurationTask,
    "VariableDurationTask": VariableDurationTask,
}
task_type_widget = widgets.Dropdown(
    options=list(task_types),
    value="FixedDurationTask",
    description="Task type:",
    disabled=False,
)


def on_change_task_type(change):
    # print("popo", change)
    if change["type"] == "change" and change["name"] == "value":
        new_value = "%s" % change["new"]
        if new_value == "FixedDurationTask":
            task_duration_widget.disabled = False
        elif new_value == "ZeroDurationTask":
            task_duration_widget.disabled = True
        elif new_value == "VariableDurationTask":
            task_duration_widget.disabled = True


task_type_widget.observe(on_change_task_type)

task_name_widget = widgets.Text(
    value="",
    placeholder="Enter task name",
    description="Task name:",
    disabled=False,
    layout={"width": "200px"},
)
is_optional_widget = widgets.Checkbox(
    value=False,
    description="Optional",
    disabled=False,
    indent=False,
    layout={"width": "200px"},
)
task_duration_widget = widgets.IntText(
    value=1, description="Duration:", disabled=False, layout={"width": "200px"}
)
task_priority_widget = widgets.IntText(
    value=1, description="Priority:", disabled=False, layout={"width": "200px"}
)
task_work_amount_widget = widgets.IntText(
    value=1, description="Work amounts:", disabled=False, layout={"width": "200px"}
)
create_task_button = widgets.Button(
    description="Create task", disabled=False, button_style="", tooltip="Click me"
)
task_output = widgets.Output()


def on_create_task_button_clicked(b):
    task_optional = is_optional_widget.value
    task_name = task_name_widget.value
    task_class = task_types[task_type_widget.value]
    if task_class == ZeroDurationTask:
        task_class(task_name)
    elif task_class == VariableDurationTask:
        task_class(
            task_name,
            # priority=task_priority_widget.value,
            # work_amount=task_work_amount_widget.value,
            optional=task_optional,
        )
    else:
        task_class(
            task_name,
            duration=task_duration_widget.value,
            # priority=task_priority_widget.value,
            # work_amount=task_work_amount_widget.value,
            optional=task_optional,
        )
    # rebuild option list for the task list
    tasks_list_of_tuples = []
    for task in pb.context.tasks:
        tasks_list_of_tuples.append((task.name, task))
    tasks_select_widget.options = tasks_list_of_tuples
    with task_output:
        print("Task", task_name, "successfully created.")


create_task_button.on_click(on_create_task_button_clicked)
task_ui = widgets.VBox(
    [
        widgets.HBox(
            [
                is_optional_widget,
                task_type_widget,
                widgets.VBox(
                    [
                        task_name_widget,
                        task_duration_widget,
                        task_priority_widget,
                        task_work_amount_widget,
                    ]
                ),
                create_task_button,
            ]
        ),
        task_output,
    ]
)
#
# Constraints UI, both tasks and resources
#
tasks_select_widget = widgets.SelectMultiple(description="Tasks:", disabled=False)
resources_select_widget = widgets.SelectMultiple(
    description="Resources:", disabled=False
)
task_constraint_types = {
    "TaskPrecedence": TaskPrecedence,
    "TasksStartSynced": TasksStartSynced,
    "TasksEndSynced": TasksEndSynced,
    "TasksDontOverlap": TasksDontOverlap,
    "TaskStartAt": TaskStartAt,
    "TaskStartAfterStrict": TaskStartAfterStrict,
    "TaskStartAfterLax": TaskStartAfterLax,
    "TaskEndAt": TaskEndAt,
    "TaskEndBeforeStrict": TaskEndBeforeStrict,
    "TaskEndBeforeLax": TaskEndBeforeLax,
    "OptionalTaskConditionSchedule": OptionalTaskConditionSchedule,
    "OptionalTasksDependency": OptionalTasksDependency,
    "ForceScheduleNOptionalTasks": ForceScheduleNOptionalTasks,
    "ScheduleNTasksInTimeIntervals": ScheduleNTasksInTimeIntervals,
}
task_constraint_type_widget = widgets.Dropdown(
    options=list(task_constraint_types),
    description="",
    layout={"width": "250px"},
    disabled=False,
)
resource_constraint_types = {
    "WorkLoad": WorkLoad,
    "ResourceUnavailable": ResourceUnavailable,
    "DistinctWorkers": DistinctWorkers,
    "SameWorkers": SameWorkers,
}
resource_constraint_type_widget = widgets.Dropdown(
    options=list(resource_constraint_types),
    description="",
    layout={"width": "250px"},
    disabled=False,
)
create_task_constraint_button = widgets.Button(
    description="Create task constraint",
    disabled=False,
    button_style="",
    tooltip="Create task constraint",
)
create_resource_constraint_button = widgets.Button(
    description="Create resource constraint",
    disabled=False,
    button_style="",
    tooltip="Create resource constraint",
)
assign_all_workers_resource_button = widgets.Button(
    description="Assign workers", disabled=False, indent=True
)
assign_resource_output = widgets.Output()


def assign_all_workers_resource_button_clicked(b) -> bool:
    # create the solver
    with assign_resource_output:
        if len(tasks_select_widget.value) != 1:
            print("Warning: select one and only one task.")
            return False
        selected_task = tasks_select_widget.value[0]
        if len(resources_select_widget.value) < 1:
            print("Warning: select at least 1 resources for SelectWorkers.")
            return False
        selected_resources = resources_select_widget.value
        # assign resources to task
        selected_task.add_required_resources(selected_resources)
        print(
            "Assign",
            ",".join([s.name for s in selected_resources]),
            "to task",
            selected_task.name,
        )
    return True


assign_all_workers_resource_button.on_click(assign_all_workers_resource_button_clicked)

assign_alternative_workers_resource_button = widgets.Button(
    description="Select 2 workers", disabled=False, indent=True
)
nb_workers_widget = widgets.IntText(
    value=2, description="NbWorkers:", disabled=False, layout={"width": "150px"}
)


def on_nb_workers_value_change(change):
    # print(change['new'])
    assign_alternative_workers_resource_button.description = (
        "Select %i workers" % change["new"]
    )


nb_workers_widget.observe(on_nb_workers_value_change, names="value")
select_worker_type_widget = widgets.Dropdown(
    options=["exact", "min", "max"],
    value="max",
    description="Type:",
    disabled=False,
    layout={"width": "150px"},
)


def assign_alternative_workers_resource_button_clicked(b) -> bool:
    # create the solver
    with assign_resource_output:
        if len(tasks_select_widget.value) != 1:
            print("Warning: select one and only one task.")
            return False
        selected_task = tasks_select_widget.value[0]
        if len(resources_select_widget.value) <= 1:
            print("Warning: select at least 2 resources for SelectWorkers.")
            return False
        selected_resources = resources_select_widget.value
        # assign resources to task
        selected_task.add_required_resource(
            SelectWorkers(
                list_of_workers=selected_resources,
                nb_workers_to_select=nb_workers_widget.value,
                kind=select_worker_type_widget.value,
            )
        )
        print(
            "Assign %i (%s) resources among"
            % (nb_workers_widget.value, select_worker_type_widget.value),
            ",".join([s.name for s in selected_resources]),
            "to task",
            selected_task.name,
        )
        return True


assign_alternative_workers_resource_button.on_click(
    assign_alternative_workers_resource_button_clicked
)


constraint_ui = widgets.VBox(
    [
        widgets.HBox(
            [
                resources_select_widget,
                resource_constraint_type_widget,
                create_resource_constraint_button,
            ]
        ),
        widgets.HBox(
            [
                tasks_select_widget,
                task_constraint_type_widget,
                create_task_constraint_button,
            ]
        ),
        widgets.HBox(
            [
                widgets.VBox(
                    [assign_all_workers_resource_button],
                    layout=widgets.Layout(border="solid 1px gray"),
                ),
                widgets.VBox(
                    [
                        assign_alternative_workers_resource_button,
                        nb_workers_widget,
                        select_worker_type_widget,
                    ],
                    layout=widgets.Layout(border="solid 1px gray"),
                ),
            ]
        ),
        assign_resource_output,
    ]
)
#
# Optimization UI
#
is_makespan_widget = widgets.Checkbox(
    value=False,
    description="MakeSpan",
    disabled=False,
    indent=False,
    layout={"width": "200px"},
)
is_flowtime_widget = widgets.Checkbox(
    value=False,
    description="Flowtime",
    disabled=False,
    indent=False,
    layout={"width": "200px"},
)
is_cost_widget = widgets.Checkbox(
    value=False,
    description="GlobalCost",
    disabled=False,
    indent=False,
    layout={"width": "200px"},
)
is_priority_widget = widgets.Checkbox(
    value=False,
    description="Priorities",
    disabled=False,
    indent=False,
    layout={"width": "200px"},
)
optimization_ui = widgets.HBox(
    [is_makespan_widget, is_flowtime_widget, is_cost_widget, is_priority_widget]
)


#
# Solver UI
#
is_debug_solver_widget = widgets.Checkbox(
    value=False,
    description="Debug",
    disabled=False,
    indent=False,
    layout={"width": "200px"},
)
is_parallel_solver_widget = widgets.Checkbox(
    value=False,
    description="Parallel",
    disabled=False,
    indent=False,
    layout={"width": "200px"},
)
max_time_widget = widgets.IntText(
    value=60, description="Max time (s):", disabled=False, layout={"width": "200px"}
)
solve_button = widgets.Button(description="Solve", disabled=False)
solve_output = widgets.Output()


def on_solve_task_button_clicked(b):
    # create the solver
    solve_output.clear_output()
    with solve_output:
        # according to the optimization, enable the related function
        if is_makespan_widget.value:
            pb.add_objective_makespan()
        if is_flowtime_widget.value:
            pb.add_objective_flowtime()
        if is_priority_widget.value:
            pb.add_objective_priorities()
        if is_cost_widget.value:
            pb.add_objective_resource_cost([pb.context.resources])
        solver = SchedulingSolver(
            pb,
            debug=is_debug_solver_widget.value,
            max_time=max_time_widget.value,
            parallel=is_parallel_solver_widget.value,
        )
        solution = solver.solve()
        # choose the gantt renderer
        print(solution)
        if pb.start_time is not None and pb.delta_time is not None:
            solution.render_gantt_plotly(render_mode="Resource")
        else:
            solution.render_gantt_matplotlib(render_mode="Resource")


solve_button.on_click(on_solve_task_button_clicked)
solver_ui = widgets.VBox(
    [
        widgets.HBox(
            [
                max_time_widget,
                widgets.VBox([is_debug_solver_widget, is_parallel_solver_widget]),
                solve_button,
            ]
        ),
        solve_output,
    ]
)



"""Resource constraints and related classes."""

# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

from typing import Optional
import uuid

from z3 import And, Implies, Int, Not, Or, Sum, Xor

from processscheduler.resource import Worker, CumulativeWorker
from processscheduler.base import _Constraint


class WorkLoad(_Constraint):
    """set a mini/maxi/exact number of slots a resource can be scheduled."""

    def __init__(
        self,
        resource,
        dict_time_intervals_and_bound,
        kind: Optional[str] = "max",
        optional: Optional[bool] = False,
    ) -> None:
        """WorkLoad constraints can be used to restrict the number tasks which are executed during a certain time period.
        The resource can be a single Worker or a CumulativeWorker.

        The list of time_intervals is a dict such as:
        [(1,20):6, (50,60):2] which means: in the interval (1,20), the resource might not use
        more than 6 slots. And no more than 2 time slots in the interval (50, 60)

        kind: optional string, default to 'max', can be 'min' or 'exact'
        """
        super().__init__(optional)

        if kind not in ["exact", "max", "min"]:
            raise ValueError("kind must either be 'exact', 'min' or 'max'")

        if isinstance(resource, Worker):
            workers = [resource]
        elif isinstance(resource, CumulativeWorker):
            workers = resource.cumulative_workers

        for time_interval in dict_time_intervals_and_bound:
            number_of_time_slots = dict_time_intervals_and_bound[time_interval]

            time_interval_lower_bound, time_interval_upper_bound = time_interval

            durations = []

            for worker in workers:
                # for this task, the logic expression is that any of its start or end must be
                # between two consecutive intervals
                for start_task_i, end_task_i in worker.get_busy_intervals():
                    # this variable allows to compute the occupation
                    # of the resource during the time interval
                    dur = Int(
                        "Overlap_%i_%i_%s"
                        % (
                            time_interval_lower_bound,
                            time_interval_upper_bound,
                            uuid.uuid4().hex[:8],
                        )
                    )
                    # prevent solutions where duration would be negative
                    self.set_assertions(dur >= 0)
                    # 4 different cases to take into account
                    cond1 = And(
                        start_task_i >= time_interval_lower_bound,
                        end_task_i <= time_interval_upper_bound,
                    )
                    asst1 = Implies(cond1, dur == end_task_i - start_task_i)
                    self.set_assertions(asst1)
                    # overlap at lower bound
                    cond2 = And(
                        start_task_i < time_interval_lower_bound,
                        end_task_i > time_interval_lower_bound,
                    )
                    asst2 = Implies(
                        cond2, dur == end_task_i - time_interval_lower_bound
                    )
                    self.set_assertions(asst2)
                    # overlap at upper bound
                    cond3 = And(
                        start_task_i < time_interval_upper_bound,
                        end_task_i > time_interval_upper_bound,
                    )
                    asst3 = Implies(
                        cond3, dur == time_interval_upper_bound - start_task_i
                    )
                    self.set_assertions(asst3)
                    # all overlap
                    cond4 = And(
                        start_task_i < time_interval_lower_bound,
                        end_task_i > time_interval_upper_bound,
                    )
                    asst4 = Implies(
                        cond4,
                        dur == time_interval_upper_bound - time_interval_lower_bound,
                    )
                    self.set_assertions(asst4)

                    # make these constraints mutual: no overlap
                    self.set_assertions(
                        Implies(Not(Or([cond1, cond2, cond3, cond4])), dur == 0)
                    )

                    # finally, store this variable in the duratins list
                    durations.append(dur)

            # workload constraint depends on the kind
            if kind == "exact":
                wl_constrt = Sum(durations) == number_of_time_slots
            elif kind == "max":
                wl_constrt = Sum(durations) <= number_of_time_slots
            elif kind == "min":
                wl_constrt = Sum(durations) >= number_of_time_slots

            self.set_assertions(wl_constrt)


class ResourceUnavailable(_Constraint):
    """set unavailablity or a resource, in terms of busy intervals"""

    def __init__(
        self, resource, list_of_time_intervals, optional: Optional[bool] = False
    ) -> None:
        """

        :param resource: the resource
        :param list_of_intervals: periods for which the resource is unavailable for any task.
        for example [(0, 2), (5, 8)]
        """
        super().__init__(optional)

        # for each interval we create a task 'UnavailableResource%i'
        if isinstance(resource, Worker):
            workers = [resource]
        elif isinstance(resource, CumulativeWorker):
            workers = resource.cumulative_workers

        for interval_lower_bound, interval_upper_bound in list_of_time_intervals:
            # add constraints on each busy interval
            for worker in workers:
                for start_task_i, end_task_i in worker.get_busy_intervals():
                    self.set_assertions(
                        Xor(
                            start_task_i >= interval_upper_bound,
                            end_task_i <= interval_lower_bound,
                        )
                    )


#
# SelectWorker specific constraints
#
class SameWorkers(_Constraint):
    """Selected workers by both AlternateWorkers are constrained to
    be the same
    """

    def __init__(
        self, alternate_workers_1, alternate_workers_2, optional: Optional[bool] = False
    ):
        super().__init__(optional)
        # we check resources in alt work 1, if it is present in
        # Select worker 2 as well, then add a constraint
        for res_work_1 in alternate_workers_1.selection_dict:
            if res_work_1 in alternate_workers_2.selection_dict:
                self.set_assertions(
                    alternate_workers_1.selection_dict[res_work_1]
                    == alternate_workers_2.selection_dict[res_work_1]
                )


class DistinctWorkers(_Constraint):
    """Selected workers by both AlternateWorkers are constrained to
    be the same
    """

    def __init__(
        self, alternate_workers_1, alternate_workers_2, optional: Optional[bool] = False
    ):
        super().__init__(optional)
        # we check resources in alt work 1, if it is present in
        # alterna worker 2 as well, then add a constraint
        for res_work_1 in alternate_workers_1.selection_dict:
            if res_work_1 in alternate_workers_2.selection_dict:
                self.set_assertions(
                    alternate_workers_1.selection_dict[res_work_1]
                    != alternate_workers_2.selection_dict[res_work_1]
                )



"""Problem definition and related classes."""

# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

from datetime import timedelta, datetime
import uuid
from typing import List, Optional, Union

from z3 import And, BoolRef, Int, Or, Sum, Implies, ArithRef

from processscheduler.base import _NamedUIDObject, is_strict_positive_integer
from processscheduler.objective import Indicator, MaximizeObjective, MinimizeObjective
from processscheduler.resource import _Resource, CumulativeWorker
from processscheduler.cost import ConstantCostPerPeriod, PolynomialCostFunction
import processscheduler.context as ps_context


class SchedulingProblem(_NamedUIDObject):
    """A scheduling problem

    :param name: the problem name, a string type
    :param horizon: an optional integer, the final instant of the timeline
    :param delta_time: an optional timedelta object
    :param start_time: an optional datetime object
    :param end_time: an optional datetime object
    :param datetime_format: an optional string

    """

    def __init__(
        self,
        name: str,
        horizon: Optional[int] = None,
        delta_time: Optional[timedelta] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
    ):
        super().__init__(name)
        # the problem context, where all will be stored
        # at creation
        self.context = ps_context.SchedulingContext()
        # set this context as global
        ps_context.main_context = self.context

        # define the horizon variable
        self.horizon = Int("horizon")
        self.fixed_horizon = False  # set to True is horizon is fixed
        if is_strict_positive_integer(horizon):  # fixed_horizon
            self.context.add_constraint(self.horizon == horizon)
            self.fixed_horizon = True
        elif horizon is not None:
            raise TypeError("horizon must either be a strict positive integer or None")

        # check time data
        if delta_time is not None and not isinstance(delta_time, timedelta):
            raise TypeError("delta_time must be a timedelta instance")
        if start_time is not None and not isinstance(start_time, datetime):
            raise TypeError("start_time must be a datetime instance")
        if end_time is not None and not isinstance(end_time, datetime):
            raise TypeError("delta_time must be a datetime instance")

        self.delta_time = delta_time
        self.start_time = start_time
        self.end_time = end_time

    def add_constraint(self, constraint: BoolRef) -> None:
        self.context.add_constraint(constraint)

    def add_constraints(self, list_of_constraints: List[BoolRef]) -> None:
        """adds constraints to the problem"""
        for cstr in list_of_constraints:
            self.context.add_constraint(cstr)

    def add_indicator_resource_cost(
        self, list_of_resources: List[_Resource]
    ) -> Indicator:
        """compute the total cost of a set of resources"""
        partial_costs = []

        def get_resource_cost(res):
            p = []
            for interv_low, interv_up in res.busy_intervals.values():
                # Constant cost per period
                if isinstance(res.cost, ConstantCostPerPeriod):
                    # res.cost(interv_up), res.cost(interv_low)
                    # or res.cost.value give the same result because the function is constant
                    period_cost = res.cost(interv_up) * (interv_up - interv_low)
                    p.append(period_cost)
                # Polynomial cost. Compute the area of the trapeze
                if isinstance(res.cost, PolynomialCostFunction):
                    period_cost = (
                        (res.cost(interv_low) + res.cost(interv_up))
                        * (interv_up - interv_low)
                        / 2
                    )
                    p.append(period_cost)
            return p

        for resource in list_of_resources:
            if isinstance(resource, CumulativeWorker):
                for res in resource.cumulative_workers:
                    partial_costs.extend(get_resource_cost(res))
            else:  # for a single worker
                partial_costs.extend(get_resource_cost(resource))

        resource_names = ",".join([resource.name for resource in list_of_resources])
        cost_indicator_variable = Sum(partial_costs)
        cost_indicator = Indicator(
            "Total Cost (%s)" % resource_names, cost_indicator_variable
        )
        return cost_indicator

    def add_indicator_resource_utilization(self, resource: _Resource) -> Indicator:
        """Compute the total utilization of a single resource.

        The percentage is rounded to an int value.
        """
        durations = []
        for interv_low, interv_up in resource.busy_intervals.values():
            durations.append(interv_up - interv_low)
        utilization = (Sum(durations) * 100) / self.horizon  # in percentage
        utilization_indicator = Indicator(
            "Utilization (%s)" % resource.name, utilization, bounds=(0, 100)
        )
        return utilization_indicator

    def maximize_indicator(self, indicator: Indicator) -> MaximizeObjective:
        """Maximize indicator"""
        return MaximizeObjective("", indicator)

    def minimize_indicator(self, indicator: Indicator) -> MinimizeObjective:
        """Minimize indicator"""
        return MinimizeObjective("", indicator)

    #
    # Optimization objectives
    #
    def add_objective_makespan(self, weight=1) -> Union[ArithRef, Indicator]:
        """makespan objective"""
        if self.fixed_horizon:
            raise ValueError(
                "Horizon constrained to be fixed, no horizon optimization possible."
            )
        MinimizeObjective("MakeSpan", self.horizon, weight)
        return self.horizon

    def add_objective_resource_utilization(
        self, resource: _Resource, weight=1
    ) -> Union[ArithRef, Indicator]:
        """Maximize resource occupation."""
        resource_utilization_indicator = self.add_indicator_resource_utilization(
            resource
        )
        MaximizeObjective("", resource_utilization_indicator, weight)
        return resource_utilization_indicator

    def add_objective_resource_cost(
        self, list_of_resources: List[_Resource], weight=1
    ) -> Union[ArithRef, Indicator]:
        """minimise the cost of selected resources"""
        cost_indicator = self.add_indicator_resource_cost(list_of_resources)
        MinimizeObjective("", cost_indicator, weight)
        return cost_indicator

    def add_objective_priorities(self, weight=1) -> Union[ArithRef, Indicator]:
        """optimize the solution such that all task with a higher
        priority value are scheduled before other tasks"""
        all_priorities = []
        for task in self.context.tasks:
            if task.optional:
                all_priorities.append(task.end * task.priority * task.scheduled)
            else:
                all_priorities.append(task.end * task.priority)
        priority_sum = Sum(all_priorities)
        priority_indicator = Indicator("PriorityTotal", priority_sum)
        MinimizeObjective("", priority_indicator, weight)
        return priority_indicator

    def add_objective_start_latest(self, weight=1) -> Union[ArithRef, Indicator]:
        """maximize the minimum start time, i.e. all the tasks
        are scheduled as late as possible"""
        mini = Int("SmallestStartTime")
        smallest_start_time = Indicator("SmallestStartTime", mini)
        smallest_start_time.add_assertion(
            Or([mini == task.start for task in self.context.tasks])
        )
        for tsk in self.context.tasks:
            smallest_start_time.add_assertion(mini <= tsk.start)
        MaximizeObjective("", smallest_start_time, weight)
        return smallest_start_time

    def add_objective_start_earliest(self, weight=1) -> Union[ArithRef, Indicator]:
        """minimize the greatest start time, i.e. tasks are schedules
        as early as possible"""
        maxi = Int("GreatestStartTime")
        greatest_start_time = Indicator("GreatestStartTime", maxi)
        greatest_start_time.add_assertion(
            Or([maxi == task.start for task in self.context.tasks])
        )
        for tsk in self.context.tasks:
            greatest_start_time.add_assertion(maxi >= tsk.start)
        MinimizeObjective("", greatest_start_time, weight)
        return greatest_start_time

    def add_objective_flowtime(self, weight=1) -> Union[ArithRef, Indicator]:
        """the flowtime is the sum of all ends, minimize. Be carful that
        it is contradictory with makespan"""
        task_ends = []
        for task in self.context.tasks:
            if task.optional:
                task_ends.append(task.end * task.scheduled)
            else:
                task_ends.append(task.end)
        flow_time_expr = Sum(task_ends)
        flow_time = Indicator("FlowTime", flow_time_expr)
        MinimizeObjective("", flow_time, weight)
        return flow_time

    def add_objective_flowtime_single_resource(
        self, resource, time_interval=None, weight=1
    ) -> Union[ArithRef, Indicator]:
        """Optimize flowtime for a single resource, for all the tasks scheduled in the
        time interval provided. Is ever no time interval is passed to the function, the
        flowtime is minimized for all the tasks scheduled in the workplan."""
        if time_interval is not None:
            lower_bound, upper_bound = time_interval
        else:
            lower_bound = 0
            upper_bound = self.horizon
        uid = uuid.uuid4().hex
        # for this resource, we look for the minimal starting time of scheduled tasks
        # as well as the maximum
        flowtime = Int("FlowtimeSingleResource%s_%s" % (resource.name, uid))

        flowtime_single_resource_indicator = Indicator(
            "FlowTime(%s:%i:%s)" % (resource.name, lower_bound, upper_bound), flowtime
        )
        # find the max end time in the time_interval
        maxi = Int(
            "GreatestTaskEndTimeInTimePeriodForResource%s_%s" % (resource.name, uid)
        )

        asst_max = []
        for task in resource.busy_intervals:
            asst_max.append(
                Implies(
                    And(task.end <= upper_bound, task.start >= lower_bound),
                    maxi == task.end,
                )
            )
        flowtime_single_resource_indicator.add_assertion(Or(asst_max))
        for task in resource.busy_intervals:
            flowtime_single_resource_indicator.add_assertion(
                Implies(
                    And(task.end <= upper_bound, task.start >= lower_bound),
                    maxi >= task.end,
                )
            )

        # and the mini
        mini = Int(
            "SmallestTaskEndTimeInTimePeriodForResource%s_%s" % (resource.name, uid)
        )

        asst_min = []
        for task in resource.busy_intervals:
            asst_min.append(
                Implies(
                    And(task.end <= upper_bound, task.start <= lower_bound),
                    mini == task.start,
                )
            )
        flowtime_single_resource_indicator.add_assertion(Or(asst_min))
        for task in resource.busy_intervals:
            flowtime_single_resource_indicator.add_assertion(
                Implies(
                    And(task.end <= upper_bound, task.start >= lower_bound),
                    mini <= task.start,
                )
            )

        # the quantity to optimize
        flowtime_single_resource_indicator.add_assertion(flowtime == maxi - mini)
        flowtime_single_resource_indicator.add_assertion(flowtime >= 0)

        MinimizeObjective("", flowtime_single_resource_indicator, weight)
        return flowtime_single_resource_indicator



# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import os
import unittest

import processscheduler as ps


def build_complex_problem(name: str, n: int) -> ps.SchedulingProblem:
    """returns a problem with n tasks and n * 3 workers"""
    problem = ps.SchedulingProblem(name)

    nb_mandatory_tasks = 3 * n
    nb_optional_tasks = n
    nb_workers = 4 * n

    mandatory_tasks = [
        ps.FixedDurationTask("mand_task%i" % i, duration=i % 8 + 1)
        for i in range(nb_mandatory_tasks)
    ]

    # n/10 optional tasks
    optional_tasks = [
        ps.FixedDurationTask("opt_task%i" % i, duration=i % 8 + 1, optional=True)
        for i in range(nb_optional_tasks)
    ]

    all_tasks = mandatory_tasks + optional_tasks

    workers = [ps.Worker("task%i" % i) for i in range(nb_workers)]

    # for each task, add three single required workers
    for i, task in enumerate(all_tasks):
        j = i + 1  # an overlap
        task.add_required_resources(workers[j - 1 : i + 3])

    return problem


def _solve_problem(problem, debug=True):
    """create a solver instance, return True if sat else False"""
    solver = ps.SchedulingSolver(problem, debug)
    solution = solver.solve()
    return solution


class TestSolver(unittest.TestCase):
    def test_schedule_single_fixed_duration_task(self) -> None:
        problem = ps.SchedulingProblem("SingleFixedDurationTaskScheduling", horizon=2)
        task = ps.FixedDurationTask("task", duration=2)

        solution = _solve_problem(problem)
        self.assertTrue(solution)
        # task should have been scheduled with start at 0
        # and end at 2
        task_solution = solution.tasks[task.name]
        self.assertEqual(task_solution.start, 0)
        self.assertEqual(task_solution.end, 2)

    def test_schedule_single_variable_duration_task(self) -> None:
        problem = ps.SchedulingProblem("SingleVariableDurationTaskScheduling")
        task = ps.VariableDurationTask("task")

        # add two constraints to set start and end
        problem.add_constraint(ps.TaskStartAt(task, 1))
        problem.add_constraint(ps.TaskEndAt(task, 4))

        solution = _solve_problem(problem)
        self.assertTrue(solution)
        # task should have been scheduled with start at 0
        # and end at 2
        task_solution = solution.tasks[task.name]
        self.assertEqual(task_solution.start, 1)
        self.assertEqual(task_solution.duration, 3)
        self.assertEqual(task_solution.end, 4)

    def test_schedule_two_fixed_duration_task_with_precedence(self) -> None:
        problem = ps.SchedulingProblem("TwoFixedDurationTasksWithPrecedence", horizon=5)
        task_1 = ps.FixedDurationTask("task1", duration=2)
        task_2 = ps.FixedDurationTask("task2", duration=3)

        # add two constraints to set start and end
        problem.add_constraint(ps.TaskStartAt(task_1, 0))
        problem.add_constraint(ps.TaskPrecedence(task_before=task_1, task_after=task_2))
        solution = _solve_problem(problem)
        self.assertTrue(solution)

        task_1_solution = solution.tasks[task_1.name]
        task_2_solution = solution.tasks[task_2.name]

        self.assertEqual(task_1_solution.start, 0)
        self.assertEqual(task_1_solution.end, 2)
        self.assertEqual(task_2_solution.start, 2)
        self.assertEqual(task_2_solution.end, 5)

    def test_schedule_single_task_single_resource(self) -> None:
        problem = ps.SchedulingProblem("SingleTaskSingleResource", horizon=7)

        task = ps.FixedDurationTask("task", duration=7)

        worker = ps.Worker("worker")

        task.add_required_resource(worker)

        solution = _solve_problem(problem)
        self.assertTrue(solution)
        # task should have been scheduled with start at 0
        # and end at 2
        task_solution = solution.tasks[task.name]

        self.assertEqual(task_solution.start, 0)
        self.assertEqual(task_solution.end, 7)
        self.assertEqual(task_solution.assigned_resources, ["worker"])

    def test_schedule_two_tasks_two_alternative_workers(self) -> None:
        problem = ps.SchedulingProblem("TwoTasksTwoSelectWorkers", horizon=4)
        # two tasks
        task_1 = ps.FixedDurationTask("task1", duration=3)
        task_2 = ps.FixedDurationTask("task2", duration=2)
        # two workers
        worker_1 = ps.Worker("worker1")
        worker_2 = ps.Worker("worker2")

        task_1.add_required_resource(ps.SelectWorkers([worker_1, worker_2], 1))
        task_2.add_required_resource(ps.SelectWorkers([worker_1, worker_2], 1))

        solution = _solve_problem(problem)
        self.assertTrue(solution)
        # each task should have one worker assigned
        task_1_solution = solution.tasks[task_1.name]
        task_2_solution = solution.tasks[task_2.name]
        self.assertEqual(len(task_1_solution.assigned_resources), 1)
        self.assertEqual(len(task_1_solution.assigned_resources), 1)
        self.assertFalse(
            task_1_solution.assigned_resources == task_2_solution.assigned_resources
        )

    def test_schedule_three_tasks_three_alternative_workers(self) -> None:
        problem = ps.SchedulingProblem("ThreeTasksThreeSelectWorkers")
        # two tasks
        task_1 = ps.FixedDurationTask("task1", duration=3)
        task_2 = ps.FixedDurationTask("task2", duration=2)
        task_3 = ps.FixedDurationTask("task3", duration=2)

        # three workers
        worker_1 = ps.Worker("worker1")
        worker_2 = ps.Worker("worker2")
        worker_3 = ps.Worker("worker3")

        all_workers = [worker_1, worker_2, worker_3]
        task_1.add_required_resource(ps.SelectWorkers(all_workers, 1))
        task_2.add_required_resource(ps.SelectWorkers(all_workers, 2))
        task_3.add_required_resource(ps.SelectWorkers(all_workers, 3))

        solution = _solve_problem(problem)
        self.assertTrue(solution)
        # each task should have one worker assigned
        self.assertEqual(len(solution.tasks[task_1.name].assigned_resources), 1)
        self.assertEqual(len(solution.tasks[task_2.name].assigned_resources), 2)
        self.assertEqual(len(solution.tasks[task_3.name].assigned_resources), 3)

    def test_alternative_workers_2(self) -> None:
        # problem
        pb_alt = ps.SchedulingProblem("AlternativeWorkerExample")

        # tasks
        t1 = ps.FixedDurationTask("t1", duration=3)
        t2 = ps.FixedDurationTask("t2", duration=2)
        t3 = ps.FixedDurationTask("t3", duration=2)
        t4 = ps.FixedDurationTask("t4", duration=2)
        t5 = ps.FixedDurationTask("t5", duration=2)

        # resource requirements
        w1 = ps.Worker("W1")
        w2 = ps.Worker("W2")
        w3 = ps.Worker("W3")
        w4 = ps.SelectWorkers([w1, w2, w3], nb_workers_to_select=1, kind="exact")
        w5 = ps.SelectWorkers([w1, w2, w3], nb_workers_to_select=2, kind="max")
        w6 = ps.SelectWorkers([w1, w2, w3], nb_workers_to_select=3, kind="min")

        # resource assignment
        t1.add_required_resource(w1)  # t1 only needs w1
        t2.add_required_resource(w2)  # t2 only needs w2
        t3.add_required_resource(w4)  # t3 needs one of w1, 2 or 3
        t4.add_required_resource(w5)  # t4 needs at most 2 of w1, w2 or 3
        t5.add_required_resource(w6)  # t5 needs at least 3 of w1, w2 or w3

        # add a makespan objective
        pb_alt.add_objective_makespan()

        # solve
        solver1 = ps.SchedulingSolver(pb_alt, debug=False)
        solution = solver1.solve()

        self.assertEqual(solution.horizon, 5)

    def test_unsat_1(self):
        problem = ps.SchedulingProblem("Unsat1")

        task = ps.FixedDurationTask("task", duration=7)

        # add two constraints to set start and end
        # impossible to satisfy both
        problem.add_constraint(ps.TaskStartAt(task, 1))
        problem.add_constraint(ps.TaskEndAt(task, 4))

        self.assertFalse(_solve_problem(problem))

    def test_solve_parallel(self):
        """a stress test with parallel mode solving"""
        problem = build_complex_problem("SolveParallel", 50)
        parallel_solver = ps.SchedulingSolver(problem, parallel=True)
        solution = parallel_solver.solve()
        self.assertTrue(solution)

    # TODO: failing test on some azure instances
    # def test_solve_max_time(self):
    #     """ a stress test which  """
    #     problem = build_complex_problem('SolveMaxTime', 1000)
    #     problem.add_objective_makespan()
    #     # 1s is not enough to solve this problem
    #     max_time_solver = ps.SchedulingSolver(problem, max_time=1)
    #     solution = max_time_solver.solve()
    #     self.assertFalse(solution)

    def test_solve_non_integer_max_time(self):
        """a stress test which"""
        problem = build_complex_problem("SolveMaxTime", 1000)
        problem.add_objective_makespan()
        # 0.5s is not enough to solve this problem
        max_time_solver = ps.SchedulingSolver(problem, max_time=0.05)
        solution = max_time_solver.solve()
        self.assertFalse(solution)

    #
    # Objectives
    #
    def test_makespan_objective(self):
        problem = build_complex_problem("SolveMakeSpanObjective", 20)
        # first look for a solution without optimization
        solution_1 = _solve_problem(problem)
        self.assertTrue(solution_1)

        horizon_without_optimization = solution_1.horizon
        # then add the objective and look for another solution
        problem.add_objective_makespan()
        # another solution
        solution_2 = _solve_problem(problem)
        self.assertTrue(solution_2)

        horizon_with_optimization = solution_2.horizon
        # horizon_with_optimization should be less than horizon_without_optimization
        self.assertLess(horizon_with_optimization, horizon_without_optimization)

    def test_flowtime_objective_big_problem(self):
        problem = build_complex_problem("SolveFlowTimeObjective", 5)  # long to compute
        problem.add_objective_flowtime()
        self.assertTrue(_solve_problem(problem))

    def test_start_latest_objective_big_problem(self):
        problem = build_complex_problem("SolveStartLatestObjective", 10)
        problem.add_objective_start_latest()
        self.assertTrue(_solve_problem(problem))

    def test_start_earliest_objective_big_problem(self):
        problem = build_complex_problem("SolveStartEarliestObjective", 10)
        problem.add_objective_start_earliest()
        self.assertTrue(_solve_problem(problem))

    def test_start_latest(self):
        problem = ps.SchedulingProblem("SolveStartLatest", horizon=51)
        # only one task, the solver should schedule a start time at 0
        task_1 = ps.FixedDurationTask("task1", duration=2)
        task_2 = ps.FixedDurationTask("task2", duration=3)

        problem.add_constraint(ps.TaskPrecedence(task_1, task_2))

        problem.add_objective_start_latest()
        solution = _solve_problem(problem)
        self.assertTrue(solution)
        # check that the task is not scheduled to start à 0
        # the only solution is 1
        self.assertEqual(solution.tasks[task_1.name].start, 51 - (3 + 2))
        self.assertEqual(solution.tasks[task_2.name].start, 51 - 3)

    def test_priorities(self):
        problem = ps.SchedulingProblem("SolvePriorities")
        task_1 = ps.FixedDurationTask("task1", duration=2, priority=1)
        task_2 = ps.FixedDurationTask("task2", duration=2, priority=10)
        task_3 = ps.FixedDurationTask("task3", duration=2, priority=100)

        problem.add_constraint(ps.TasksDontOverlap(task_1, task_2))
        problem.add_constraint(ps.TasksDontOverlap(task_2, task_3))
        problem.add_constraint(ps.TasksDontOverlap(task_1, task_3))

        problem.add_objective_priorities()

        # set debug to False because assert_and_track
        # does not properly handles optimization
        solution = _solve_problem(problem, debug=False)
        self.assertTrue(solution)
        # check that the task is not scheduled to start à 0
        # the only solution is 1
        self.assertLess(
            solution.tasks[task_3.name].start, solution.tasks[task_2.name].start
        )
        self.assertLess(
            solution.tasks[task_2.name].start, solution.tasks[task_1.name].start
        )

    #
    # Logical Operators
    #
    def test_operator_not(self):
        problem = ps.SchedulingProblem("OperatorNot", horizon=4)
        # only one task, the solver should schedule a start time at 0
        task_1 = ps.FixedDurationTask("task1", duration=3)

        problem.add_constraint(ps.not_(ps.TaskStartAt(task_1, 0)))
        solution = _solve_problem(problem)
        self.assertTrue(solution)
        # check that the task is not scheduled to start à 0
        # the only solution is 1
        self.assertTrue(solution.tasks[task_1.name].start == 1)

    def test_operator_not_and(self):
        problem = ps.SchedulingProblem("OperatorNotAnd", horizon=4)
        # only one task, the solver should schedule a start time at 0
        task_1 = ps.FixedDurationTask("task1", duration=2)
        # problem.add_task(task_1)
        problem.add_constraint(
            ps.and_(
                [ps.not_(ps.TaskStartAt(task_1, 0)), ps.not_(ps.TaskStartAt(task_1, 1))]
            )
        )
        solution = _solve_problem(problem)
        self.assertTrue(solution)
        # the only solution is to start at 2
        self.assertTrue(solution.tasks[task_1.name].start == 2)

    #
    # Implication
    #
    def test_implies(self):
        problem = ps.SchedulingProblem("Implies", horizon=6)
        # only one task, the solver should schedule a start time at 0
        task_1 = ps.FixedDurationTask("task1", duration=2)
        task_2 = ps.FixedDurationTask("task2", duration=2)
        problem.add_constraint(ps.TaskStartAt(task_1, 1))
        problem.add_constraint(
            ps.implies(task_1.start == 1, [ps.TaskStartAt(task_2, 4)])
        )
        solution = _solve_problem(problem)
        self.assertTrue(solution)
        # the only solution is to start at 2
        self.assertTrue(solution.tasks[task_1.name].start == 1)
        self.assertTrue(solution.tasks[task_2.name].start == 4)

    #
    # If then else
    #
    def test_if_then_else(self):
        problem = ps.SchedulingProblem("IfThenElse", horizon=6)
        # only one task, the solver should schedule a start time at 0
        task_1 = ps.FixedDurationTask("task1", duration=2)
        task_2 = ps.FixedDurationTask("task2", duration=2)
        problem.add_constraint(ps.TaskStartAt(task_1, 1))
        problem.add_constraint(
            ps.if_then_else(
                task_1.start == 0,  # this condition is False
                [ps.TaskStartAt(task_2, 4)],  # assertion not set
                [ps.TaskStartAt(task_2, 2)],
            )
        )  # this one
        solution = _solve_problem(problem)
        self.assertTrue(solution)
        # the only solution is to start at 2
        self.assertTrue(solution.tasks[task_1.name].start == 1)
        self.assertTrue(solution.tasks[task_2.name].start == 2)

    #
    # Find other solution
    #
    def test_find_another_solution(self):
        problem = ps.SchedulingProblem("FindAnotherSolution", horizon=6)
        solutions = []

        task_1 = ps.FixedDurationTask("task1", duration=2)
        solver = ps.SchedulingSolver(problem)
        solution = solver.solve()

        while solution:
            solutions.append(solution.tasks[task_1.name].start)
            solution = solver.find_another_solution(task_1.start)
        # there should be 5 solutions
        self.assertEqual(solutions, [0, 1, 2, 3, 4])

    def test_find_another_solution_solve_before(self):
        problem = ps.SchedulingProblem("FindAnotherSolutionSolveBefore", horizon=6)

        task_1 = ps.FixedDurationTask("task1", duration=2)
        solver = ps.SchedulingSolver(problem)
        result = solver.find_another_solution(
            task_1.start
        )  # error, first have to solve
        self.assertFalse(result)

    #
    # Total work_amount, resource productivity
    #
    def test_work_amount_1(self):
        problem = ps.SchedulingProblem("WorkAmount")

        task_1 = ps.VariableDurationTask("task1", work_amount=11)
        # create one worker with a productivity of 2
        worker_1 = ps.Worker("Worker1", productivity=2)
        task_1.add_required_resource(worker_1)
        # solve
        solution = _solve_problem(problem)
        self.assertTrue(solution)
        # the expected duration for task 1 is 6
        self.assertEqual(solution.tasks[task_1.name].duration, 6)

    def test_work_amount_2(self):
        # try the same problem than above, but with one more resource
        # check that the task duration is lower
        problem = ps.SchedulingProblem("WorkAmount", horizon=4)

        task_1 = ps.VariableDurationTask("task1", work_amount=11)
        # create two workers
        worker_1 = ps.Worker("Worker1", productivity=2)
        worker_2 = ps.Worker("Worker2", productivity=3)
        task_1.add_required_resources([worker_1, worker_2])
        # solve
        self.assertTrue(_solve_problem(problem))

    #
    # Import/export
    #
    def test_export_to_smt2(self):
        problem = build_complex_problem("SolveExportToSMT2", 50)
        solver = ps.SchedulingSolver(problem)
        solution = _solve_problem(problem)
        self.assertTrue(solution)
        solver.export_to_smt2("complex_problem.smt2")
        self.assertTrue(os.path.isfile("complex_problem.smt2"))

    def test_export_solution_to_json(self):
        problem = build_complex_problem("SolutionExportToJson", 50)
        solution = _solve_problem(problem)
        self.assertTrue(solution)
        solution.to_json_string()

    #
    # Resource constraints
    #
    def test_all_same_distinct_workers(self):
        pb = ps.SchedulingProblem("AllSameDistinctWorkers")

        task_1 = ps.FixedDurationTask("task1", duration=2)
        task_2 = ps.FixedDurationTask("task2", duration=2)
        task_3 = ps.FixedDurationTask("task3", duration=2)
        task_4 = ps.FixedDurationTask("task4", duration=2)

        worker_1 = ps.Worker("John")
        worker_2 = ps.Worker("Bob")

        res_for_t1 = ps.SelectWorkers([worker_1, worker_2], 1)
        res_for_t2 = ps.SelectWorkers([worker_1, worker_2], 1)
        res_for_t3 = ps.SelectWorkers([worker_1, worker_2], 1)
        res_for_t4 = ps.SelectWorkers([worker_1, worker_2], 1)

        task_1.add_required_resource(res_for_t1)
        task_2.add_required_resource(res_for_t2)
        task_3.add_required_resource(res_for_t3)
        task_4.add_required_resource(res_for_t4)

        c = ps.SameWorkers(res_for_t1, res_for_t2)
        d = ps.SameWorkers(res_for_t3, res_for_t4)
        e = ps.DistinctWorkers(res_for_t2, res_for_t4)
        pb.add_constraints([c, d, e])

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertEqual(solution.horizon, 4)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import unittest

import processscheduler as ps


class TestDynamicResource(unittest.TestCase):
    def test_dynamic_1(self) -> None:
        pb = ps.SchedulingProblem("DynamicTest1")
        task_1 = ps.FixedDurationTask("task1", duration=10, work_amount=10)
        task_2 = ps.FixedDurationTask("task2", duration=5)
        task_3 = ps.FixedDurationTask("task3", duration=5)

        pb.add_constraint(ps.TaskStartAt(task_3, 0))
        pb.add_constraint(ps.TaskEndAt(task_2, 10))

        worker_1 = ps.Worker("Worker1", productivity=1)
        worker_2 = ps.Worker("Worker2", productivity=1)

        task_1.add_required_resources([worker_1, worker_2], dynamic=True)
        task_2.add_required_resource(worker_1)
        task_3.add_required_resource(worker_2)

        pb.add_objective_makespan()

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertEqual(solution.horizon, 10)

    def test_non_dynamic_1(self) -> None:
        # same test as previously
        # but the task_1 workers are non dynamic.
        # so the horizon must be 20.
        pb = ps.SchedulingProblem("NonDynamicTest1")
        task_1 = ps.FixedDurationTask("task1", duration=10, work_amount=10)
        task_2 = ps.FixedDurationTask("task2", duration=5)
        task_3 = ps.FixedDurationTask("task3", duration=5)

        pb.add_constraint(ps.TaskStartAt(task_3, 0))
        pb.add_constraint(ps.TaskEndAt(task_2, 10))

        worker_1 = ps.Worker("Worker1", productivity=1)
        worker_2 = ps.Worker("Worker2", productivity=1)

        task_1.add_required_resources([worker_1, worker_2])  # dynamic False by default
        task_2.add_required_resource(worker_1)
        task_3.add_required_resource(worker_2)

        pb.add_objective_makespan()

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertEqual(solution.horizon, 20)


if __name__ == "__main__":
    unittest.main()



import processscheduler as ps

problem = ps.SchedulingProblem("Excavator7:MultiObjectiveOptimization")

# three tasks
dig_small_hole = ps.VariableDurationTask("DigSmallHole", work_amount=3, priority=1)
dig_medium_hole = ps.VariableDurationTask("DigMediumHole", work_amount=7, priority=10)
dig_huge_hole = ps.VariableDurationTask("DigHugeHole", work_amount=15, priority=100)

# two workers
small_exc = ps.Worker(
    "SmallExcavator", productivity=4, cost=ps.ConstantCostPerPeriod(5)
)
medium_ex = ps.Worker(
    "MediumExcavator", productivity=6, cost=ps.ConstantCostPerPeriod(10)
)

dig_small_hole.add_required_resource(
    ps.SelectWorkers([small_exc, medium_ex], 1, kind="min")
)
dig_medium_hole.add_required_resource(
    ps.SelectWorkers([small_exc, medium_ex], 1, kind="min")
)
dig_huge_hole.add_required_resource(
    ps.SelectWorkers([small_exc, medium_ex], 1, kind="min")
)

problem.add_objective_makespan()
problem.add_objective_resource_cost([small_exc, medium_ex])
problem.add_objective_priorities()

solver = ps.SchedulingSolver(problem, optimize_priority="pareto")
solution = solver.solve()
solution.render_gantt_matplotlib()



# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import unittest

import processscheduler as ps
import processscheduler.context as ps_context


def new_problem_or_clear() -> None:
    """clear the current context. If no context is defined,
    create a SchedulingProject object"""
    if ps_context.main_context is None:
        ps.SchedulingProblem("NewProblem")
    else:
        ps_context.main_context.clear()


class TestFeatures(unittest.TestCase):
    def test_clear_context(self) -> None:
        ps_context.main_context = None
        new_problem_or_clear()
        self.assertIsInstance(ps_context.main_context, ps.SchedulingContext)

    def test_create_problem_with_horizon(self) -> None:
        pb = ps.SchedulingProblem("ProblemWithHorizon", horizon=10)
        self.assertIsInstance(pb, ps.SchedulingProblem)
        with self.assertRaises(TypeError):
            ps.SchedulingProblem(4)  # name not string
        with self.assertRaises(TypeError):
            ps.SchedulingProblem("NullIntegerHorizon", horizon=0)
        with self.assertRaises(TypeError):
            ps.SchedulingProblem("FloatHorizon", horizon=3.5)
        with self.assertRaises(TypeError):
            ps.SchedulingProblem("NegativeIntegerHorizon", horizon=-2)

    def test_create_problem_without_horizon(self) -> None:
        pb = ps.SchedulingProblem("ProblemWithoutHorizon")
        self.assertIsInstance(pb, ps.SchedulingProblem)

    #
    # Workers
    #
    def test_create_worker(self) -> None:
        new_problem_or_clear()
        worker = ps.Worker("wkr")
        self.assertIsInstance(worker, ps.Worker)
        with self.assertRaises(TypeError):
            ps.Worker("WorkerNegativeIntProductivity", productivity=-3)
        with self.assertRaises(TypeError):
            ps.Worker("WorkerFloatProductivity", productivity=3.14)

    def test_create_select_workers(self) -> None:
        new_problem_or_clear()
        worker_1 = ps.Worker("wkr_1")
        worker_2 = ps.Worker("wkr_2")
        worker_3 = ps.Worker("wkr_3")
        single_alternative_workers = ps.SelectWorkers([worker_1, worker_2], 1)
        self.assertIsInstance(single_alternative_workers, ps.SelectWorkers)
        double_alternative_workers = ps.SelectWorkers([worker_1, worker_2, worker_3], 2)
        self.assertIsInstance(double_alternative_workers, ps.SelectWorkers)

    def test_select_worker_wrong_number_of_workers(self) -> None:
        new_problem_or_clear()
        worker_1 = ps.Worker("wkr_1")
        worker_2 = ps.Worker("wkr_2")
        ps.SelectWorkers([worker_1, worker_2], 2)
        ps.SelectWorkers([worker_1, worker_2], 1)
        with self.assertRaises(ValueError):
            ps.SelectWorkers([worker_1, worker_2], 3)
        with self.assertRaises(TypeError):
            ps.SelectWorkers([worker_1, worker_2], -1)

    def test_select_worker_bad_type(self) -> None:
        new_problem_or_clear()
        worker_1 = ps.Worker("wkr_1")
        self.assertIsInstance(worker_1, ps.Worker)
        worker_2 = ps.Worker("wkr_2")
        with self.assertRaises(ValueError):
            ps.SelectWorkers([worker_1, worker_2], 1, kind="ee")

    def test_worker_same_name(self) -> None:
        new_problem_or_clear()
        worker_1 = ps.Worker("wkr_1")
        self.assertIsInstance(worker_1, ps.Worker)
        with self.assertRaises(ValueError):
            ps.Worker("wkr_1")

    #
    # Boolean operators
    #
    def test_operator_not_(self) -> None:
        new_problem_or_clear()
        t_1 = ps.FixedDurationTask("t1", duration=2)
        not_constraint = ps.not_(ps.TaskStartAt(t_1, 1))
        self.assertIsInstance(not_constraint, ps.BoolRef)

    def test_operator_or_(self) -> None:
        new_problem_or_clear()
        t_1 = ps.FixedDurationTask("t1", duration=2)
        or_constraint = ps.or_([ps.TaskStartAt(t_1, 1), ps.TaskStartAt(t_1, 2)])
        self.assertIsInstance(or_constraint, ps.BoolRef)

    def test_operator_xor_(self) -> None:
        new_problem_or_clear()
        t_1 = ps.FixedDurationTask("t1", duration=2)
        xor_constraint = ps.xor_([ps.TaskStartAt(t_1, 1), ps.TaskStartAt(t_1, 2)])
        self.assertIsInstance(xor_constraint, ps.BoolRef)

    def test_operator_xor_2(self) -> None:
        new_problem_or_clear()
        t_1 = ps.FixedDurationTask("t1", duration=2)
        with self.assertRaises(TypeError):
            ps.xor_(
                [ps.TaskStartAt(t_1, 1), ps.TaskStartAt(t_1, 2), ps.TaskStartAt(t_1, 3)]
            )

    def test_operator_and_(self) -> None:
        new_problem_or_clear()
        t_1 = ps.FixedDurationTask("t1", duration=2)
        and_constraint = ps.and_(
            [ps.TaskStartAfterLax(t_1, 1), ps.TaskEndBeforeLax(t_1, 7)]
        )
        self.assertIsInstance(and_constraint, ps.BoolRef)

    def test_nested_boolean_operators(self) -> None:
        new_problem_or_clear()
        t_1 = ps.VariableDurationTask("t1")
        or_constraint_1 = ps.or_([ps.TaskStartAt(t_1, 1), ps.TaskStartAt(t_1, 2)])
        or_constraint_2 = ps.or_([ps.TaskStartAt(t_1, 4), ps.TaskStartAt(t_1, 5)])
        and_constraint = ps.and_([or_constraint_1, or_constraint_2])
        self.assertIsInstance(and_constraint, ps.BoolRef)

    def test_add_constraint(self) -> None:
        pb = ps.SchedulingProblem("AddConstraint")
        t_1 = ps.FixedDurationTask("t1", duration=2)
        or_constraint = ps.or_([ps.TaskStartAt(t_1, 1), ps.TaskStartAt(t_1, 2)])
        not_constraint = ps.not_(ps.TaskEndAt(t_1, 5))
        self.assertIsInstance(or_constraint, ps.BoolRef)
        self.assertIsInstance(not_constraint, ps.BoolRef)
        pb.add_constraint(or_constraint)
        pb.add_constraint(not_constraint)

    #
    # Implies
    #
    def test_implies(self) -> None:
        new_problem_or_clear()
        t_1 = ps.FixedDurationTask("t1", 2)
        t_2 = ps.FixedDurationTask("t2", 2)
        implies_constraint = ps.implies(t_1.start == 1, [ps.TaskStartAt(t_2, 3)])
        self.assertIsInstance(implies_constraint, ps.BoolRef)

    #
    # If/Then/Else
    #
    def test_if_then_else(self) -> None:
        new_problem_or_clear()
        t_1 = ps.FixedDurationTask("t1", 2)
        t_2 = ps.FixedDurationTask("t2", 2)
        ite_constraint = ps.if_then_else(
            t_1.start == 1,  # condition
            [ps.TaskStartAt(t_2, 3)],  # then
            [ps.TaskStartAt(t_2, 6)],
        )  # else
        self.assertIsInstance(ite_constraint, ps.BoolRef)

    #
    # Indicators
    #
    def test_create_indicator(self) -> None:
        pb = ps.SchedulingProblem("CreateIndicator", horizon=3)
        i_1 = ps.Indicator("SquareHorizon", pb.horizon ** 2)  # ArithRef
        self.assertIsInstance(i_1, ps.Indicator)
        i_2 = ps.Indicator("IsLooooong ?", pb.horizon > 1000)  # BoolRef
        self.assertIsInstance(i_2, ps.Indicator)
        with self.assertRaises(TypeError):
            ps.Indicator("foo", 4)

    #
    # Print _NamedUIDObject
    #
    def test_print_objects(self) -> None:
        new_problem_or_clear()
        t1 = ps.FixedDurationTask("task_1", duration=1)
        t2 = ps.VariableDurationTask("task_2")
        worker_1 = ps.Worker("W1")
        self.assertTrue("task_1" in "%s" % t1)
        self.assertTrue("task_2" in "%s" % t2)
        self.assertTrue("W1" in "%s" % worker_1)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime, timedelta
import unittest

import processscheduler as ps


class TestDatetime(unittest.TestCase):
    def test_datetime_1(self):
        """take the single task/single resource and display output"""
        problem = ps.SchedulingProblem(
            "DateTimeBase",
            horizon=7,
            delta_time=timedelta(minutes=15),
            start_time=datetime.now(),
        )
        task = ps.FixedDurationTask("task", duration=7)
        # problem.add_task(task)
        worker = ps.Worker("worker")
        # problem.add_resource(worker)
        task.add_required_resource(worker)
        solver = ps.SchedulingSolver(problem)
        solution = solver.solve()
        self.assertTrue(solution)
        print(solution)

    def test_datetime_time(self):
        """take the single task/single resource and display output"""
        problem = ps.SchedulingProblem(
            "DateTimeBase", horizon=7, delta_time=timedelta(minutes=15)
        )
        task = ps.FixedDurationTask("task", duration=7)
        # problem.add_task(task)
        worker = ps.Worker("worker")
        # problem.add_resource(worker)
        task.add_required_resource(worker)
        solver = ps.SchedulingSolver(problem)
        solution = solver.solve()
        self.assertTrue(solution)
        print(solution)

    def test_datetime_export_to_json(self):
        problem = ps.SchedulingProblem(
            "DateTimeJson", delta_time=timedelta(hours=1), start_time=datetime.now()
        )
        task = ps.FixedDurationTask("task", duration=7)
        # problem.add_task(task)
        worker = ps.Worker("worker")
        # problem.add_resource(worker)
        task.add_required_resource(worker)
        solver = ps.SchedulingSolver(problem)
        solution = solver.solve()
        self.assertTrue(solution)
        solution.to_json_string()


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import os
import unittest
from datetime import datetime, timedelta

import processscheduler as ps


class TestGantt(unittest.TestCase):
    def test_gantt_matplotlib_base(self):
        """take the single task/single resource and display output"""
        problem = ps.SchedulingProblem("RenderSolution", horizon=7)
        task = ps.FixedDurationTask("task", duration=7)
        # problem.add_task(task)
        worker = ps.Worker("worker")
        # problem.add_resource(worker)
        task.add_required_resource(worker)
        solver = ps.SchedulingSolver(problem)
        solution = solver.solve()
        self.assertTrue(solution)

        # display solution, using both ascii or matplotlib
        solution.render_gantt_matplotlib(
            render_mode="Resource",
            show_plot=False,
            fig_filename="test_render_resources_matplotlib.svg",
        )
        solution.render_gantt_matplotlib(
            render_mode="Task",
            show_plot=False,
            fig_filename="test_render_tasks_matplotlib.svg",
        )
        self.assertTrue(os.path.isfile("test_render_resources_matplotlib.svg"))
        self.assertTrue(os.path.isfile("test_render_tasks_matplotlib.svg"))

    def test_gantt_matplotlib_indicator(self):
        problem = ps.SchedulingProblem("GanttIndicator", horizon=10)

        t_1 = ps.FixedDurationTask("T1", duration=5)
        ps.FixedDurationTask("T2", duration=5)  # task with no resource
        worker_1 = ps.Worker("Worker1")
        t_1.add_required_resource(worker_1)

        problem.add_indicator_resource_utilization(worker_1)

        solution = ps.SchedulingSolver(problem).solve()

        self.assertTrue(solution)
        # display solution, using both ascii or matplotlib
        solution.render_gantt_matplotlib(render_mode="Resource", show_plot=False)

    def test_gantt_matplotlib_navaible_resource(self):
        pb = ps.SchedulingProblem("GanttResourceUnavailable", horizon=10)
        task_1 = ps.FixedDurationTask("task1", duration=3)
        worker_1 = ps.Worker("Worker1")
        task_1.add_required_resource(worker_1)
        c1 = ps.ResourceUnavailable(worker_1, [(1, 3)])
        pb.add_constraint(c1)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        solution.render_gantt_matplotlib(render_mode="Resource", show_plot=False)

    def test_gantt_matplotlib_no_resource(self):
        pb = ps.SchedulingProblem("GanttNoResource", horizon=10)
        ps.FixedDurationTask("task1", duration=3)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        solution.render_gantt_matplotlib(render_mode="Resource", show_plot=False)

    def test_gantt_matplotlib_resource_unavailable(self) -> None:
        pb = ps.SchedulingProblem("GanttResourceUnavailable1", horizon=10)
        task_1 = ps.FixedDurationTask("task1", duration=3)
        worker_1 = ps.Worker("Worker1")
        task_1.add_required_resource(worker_1)
        c1 = ps.ResourceUnavailable(worker_1, [(1, 3)])
        pb.add_constraint(c1)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        solution.render_gantt_matplotlib(show_plot=False)

    def test_gantt_matplotlib_zero_duration_task(self):
        pb = ps.SchedulingProblem("GanttZeroDuration", horizon=10)
        ps.ZeroDurationTask("task1")
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        solution.render_gantt_matplotlib(show_plot=False)

    def test_gantt_matplotlib_real_date_1(self):
        pb = ps.SchedulingProblem(
            "GanttMatplotLibRealDate1",
            horizon=10,
            start_time=datetime.now(),
            delta_time=timedelta(minutes=15),
        )
        task_1 = ps.FixedDurationTask("task1", duration=3)
        worker_1 = ps.Worker("Worker1")
        task_1.add_required_resource(worker_1)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        solution.render_gantt_matplotlib(render_mode="Task", show_plot=False)
        solution.render_gantt_matplotlib(render_mode="Resource", show_plot=False)

    def test_gantt_matplotlib_real_date_no_start_time(self):
        pb = ps.SchedulingProblem(
            "GanttMatplotLibRealDateNoStartTime",
            horizon=10,
            delta_time=timedelta(hours=2),
        )
        task_1 = ps.FixedDurationTask("task1", duration=3)
        worker_1 = ps.Worker("Worker1")
        task_1.add_required_resource(worker_1)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        solution.render_gantt_matplotlib(render_mode="Task", show_plot=False)
        solution.render_gantt_matplotlib(render_mode="Resource", show_plot=False)

    def test_gantt_matplotlib_wrong_render_mode(self):
        pb = ps.SchedulingProblem("GanttWrongRenderMode", horizon=10)
        task_1 = ps.FixedDurationTask("task1", duration=3)
        worker_1 = ps.Worker("Worker1")
        task_1.add_required_resource(worker_1)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        with self.assertRaises(ValueError):
            solution.render_gantt_matplotlib(render_mode="foo", show_plot=False)

    def test_gantt_plotly_base(self):
        """take the single task/single resource and display output"""
        problem = ps.SchedulingProblem("RenderSolutionPlotly", horizon=7)
        task = ps.FixedDurationTask("task", duration=7)
        # problem.add_task(task)
        worker = ps.Worker("worker")
        # problem.add_resource(worker)
        task.add_required_resource(worker)
        solver = ps.SchedulingSolver(problem)
        solution = solver.solve()
        self.assertTrue(solution)

        # display solution, using both ascii or matplotlib
        solution.render_gantt_plotly(
            render_mode="Resource",
            show_plot=False,
            sort="Resource",
            fig_filename="test_render_resources_plotly.svg",
        )
        solution.render_gantt_plotly(
            render_mode="Task",
            show_plot=False,
            sort="Task",
            fig_filename="test_render_tasks_plotly.svg",
        )
        solution.render_gantt_plotly(
            render_mode="Resource",
            show_plot=False,
            sort="Start",
            html_filename="test_render_resources_plotly.html",
        )
        solution.render_gantt_plotly(
            render_mode="Task",
            show_plot=False,
            sort="Finish",
            html_filename="test_render_tasks_plotly.html",
        )
        self.assertTrue(os.path.isfile("test_render_resources_plotly.svg"))
        self.assertTrue(os.path.isfile("test_render_tasks_plotly.svg"))
        self.assertTrue(os.path.isfile("test_render_resources_plotly.html"))
        self.assertTrue(os.path.isfile("test_render_tasks_plotly.html"))

    def test_gantt_plotly_with_indicators_figsize(self):
        problem = ps.SchedulingProblem("GanttIndicator", horizon=10)

        t_1 = ps.FixedDurationTask("T1", duration=5)
        ps.FixedDurationTask("T2", duration=5)  # task with no resource
        worker_1 = ps.Worker("Worker1")
        t_1.add_required_resource(worker_1)

        problem.add_indicator_resource_utilization(worker_1)

        solution = ps.SchedulingSolver(problem).solve()

        self.assertTrue(solution)
        # display solution, using both ascii or matplotlib
        solution.render_gantt_plotly(
            render_mode="Resource", show_plot=False, fig_size=(400, 300)
        )

    def test_gantt_plotly_raise_wrong_type(self):
        problem = ps.SchedulingProblem("GanttIndicator", horizon=10)

        t_1 = ps.FixedDurationTask("T1", duration=5)
        ps.FixedDurationTask("T2", duration=5)  # task with no resource
        worker_1 = ps.Worker("Worker1")
        t_1.add_required_resource(worker_1)

        solution = ps.SchedulingSolver(problem).solve()

        self.assertTrue(solution)
        # display solution, using both ascii or matplotlib
        with self.assertRaises(ValueError):
            solution.render_gantt_plotly(render_mode="foo")


if __name__ == "__main__":
    unittest.main()



from setuptools import find_packages, setup

DESCRIPTION = (
    "A package to solve resource-constrained scheduling problems using SAT/SMT."
)

CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "Intended Audience :: Manufacturing",
    "Intended Audience :: Financial and Insurance Industry",
    "Intended Audience :: Healthcare Industry",
    "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    "Natural Language :: English",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3",
    "Topic :: Office/Business :: Scheduling",
    "Topic :: Software Development :: Libraries",
    "Typing :: Typed",
]

setup(
    name="ProcessScheduler",
    version="0.7.1",
    description=DESCRIPTION,
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/tpaviot/ProcessScheduler",
    author="Thomas Paviot",
    author_email="tpaviot@gmail.com",
    license="GPLv3",
    platforms="Platform Independent",
    packages=find_packages(),
    install_requires=["z3-solver==4.8.10.0"],
    classifiers=CLASSIFIERS,
    zip_safe=True,
)



# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import unittest

import processscheduler as ps


class TestOptionalConstraint(unittest.TestCase):
    def test_optional_constraint_start_at_1(self) -> None:
        pb = ps.SchedulingProblem("OptionalTaskStartAt1", horizon=6)
        task_1 = ps.FixedDurationTask("task1", duration=3)
        # the following tasks should conflict if they are mandatory
        pb.add_constraint(ps.TaskStartAt(task_1, 1, optional=True))
        pb.add_constraint(ps.TaskStartAt(task_1, 2, optional=True))
        pb.add_constraint(ps.TaskEndAt(task_1, 3, optional=True))

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)

    def test_force_apply_n_optional_constraints(self) -> None:
        pb = ps.SchedulingProblem("OptionalTaskStartAt1", horizon=6)
        task_1 = ps.FixedDurationTask("task1", duration=3)
        # the following tasks should conflict if they are mandatory
        cstr1 = ps.TaskStartAt(task_1, 1, optional=True)
        cstr2 = ps.TaskStartAt(task_1, 2, optional=True)
        cstr3 = ps.TaskEndAt(task_1, 3, optional=True)
        # force to apply exactly one constraint
        cstr4 = ps.ForceApplyNOptionalConstraints([cstr1, cstr2, cstr3], 1)
        # if we force 2 constraints, there should not be any solution

        pb.add_constraints([cstr1, cstr2, cstr3, cstr4])

        solver = ps.SchedulingSolver(pb)

        solution = solver.solve()
        self.assertTrue(solution)


if __name__ == "__main__":
    unittest.main()



"""Solver definition and related classes/functions."""

# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import random
import time
from typing import Optional, Union
import uuid
import warnings

from z3 import Int, Solver, SolverFor, Sum, unsat, ArithRef, unknown, set_option, Or

from processscheduler.objective import MaximizeObjective, MinimizeObjective, Indicator
from processscheduler.solution import SchedulingSolution, TaskSolution, ResourceSolution

#
# Util function
#
def _calc_parabola_vertex(vector_x, vector_y):
    """Adapted from http://chris35wills.github.io/parabola_python/
    Return a, b, c such as points 1, 2, 3 satisfies
    y = ax**2+bx+c
    """
    x1, x2, x3 = vector_x
    y1, y2, y3 = vector_y
    denom = (x1 - x2) * (x1 - x3) * (x2 - x3)
    a = (x3 * (y2 - y1) + x2 * (y1 - y3) + x1 * (y3 - y2)) / denom
    b = (x3 * x3 * (y1 - y2) + x2 * x2 * (y3 - y1) + x1 * x1 * (y2 - y3)) / denom
    c = (
        x2 * x3 * (x2 - x3) * y1 + x3 * x1 * (x3 - x1) * y2 + x1 * x2 * (x1 - x2) * y3
    ) / denom
    return a, b, c


#
# Solver class definition
#
class SchedulingSolver:
    """A solver class"""

    def __init__(
        self,
        problem,
        debug: Optional[bool] = False,
        max_time: Optional[int] = 10,
        parallel: Optional[bool] = False,
        random_seed=False,
        logics=None,
        verbosity=0,
    ):
        """Scheduling Solver

        debug: True or False, False by default
        max_time: time in seconds, 60 by default
        parallel: True to enable mutlthreading, False by default
        """
        self._problem = problem
        self.problem_context = problem.context
        self.debug = debug
        # objectives list
        self.objective = None  # the list of all objectives defined in this problem
        # set_option('smt.arith.auto_config_simplex', True)
        if debug:
            set_option("verbose", 2)
        else:
            set_option("verbose", verbosity)

        if random_seed:
            set_option("sat.random_seed", random.randint(1, 1e4))
            set_option("smt.random_seed", random.randint(1, 1e4))
        else:
            set_option("sat.random_seed", 0)
            set_option("smt.random_seed", 0)

        # set timeout
        self.max_time = max_time  # in seconds
        set_option("timeout", int(self.max_time * 1000))  # in milliseconds

        # create the solver
        print("Solver type:\n===========")

        # check if the problem is an optimization problem
        self.is_not_optimization_problem = len(self.problem_context.objectives) == 0
        self.is_optimization_problem = len(self.problem_context.objectives) > 0
        self.is_multi_objective_optimization_problem = (
            len(self.problem_context.objectives) > 1
        )
        # the Optimize() solver is used only in the case of a mutli-optimization
        # problem. This enables to choose the priority method.
        # in the case of a single objective optimization, the Optimize() solver
        # apperas to be less robust than the basic Solver(). The
        # incremental solver is then used.

        # see this url for a documentation about logics
        # http://smtlib.cs.uiowa.edu/logics.shtml
        if logics is None:
            self._solver = Solver()
            print("\t-> Standard SAT/SMT solver")
        else:
            self._solver = SolverFor(logics)
            print("\t-> SMT solver using logics", logics)
        if debug:
            set_option(unsat_core=True)

        # self._solver.set("sat.cardinality.solver", True)
        # self._solver.set("sat.core.minimize", True)
        # self._solver.set("sat.core.minimize_partial", True)
        if parallel:
            set_option("parallel.enable", True)  # enable parallel computation

        # add all tasks assertions to the solver
        for task in self.problem_context.tasks:
            self.add_constraint(task.get_assertions())
            self.add_constraint(task.end <= self._problem.horizon)

        # then process tasks constraints
        for constraint in self.problem_context.constraints:
            self.add_constraint(constraint)

        # process resources requirements
        for ress in self.problem_context.resources:
            self.add_constraint(ress.get_assertions())

        # process resource intervals
        for ress in self.problem_context.resources:
            busy_intervals = ress.get_busy_intervals()
            nb_intervals = len(busy_intervals)
            for i in range(nb_intervals):
                start_task_i, end_task_i = busy_intervals[i]
                for k in range(i + 1, nb_intervals):
                    start_task_k, end_task_k = busy_intervals[k]
                    self.add_constraint(
                        Or(start_task_k >= end_task_i, start_task_i >= end_task_k)
                    )

        # process indicators
        for indic in self.problem_context.indicators:
            self.add_constraint(indic.get_assertions())

        self.process_work_amount()

        if self.is_optimization_problem:
            self.create_objective()

        # each time the solver is called, the current_solution is stored
        self.current_solution = None

    def add_constraint(self, cstr) -> bool:
        # set the method to use to add constraints
        # in debug mode this is assert_and_track, to be able to trace
        # unsat core, in regular mode this is the add function
        if self.debug:
            if isinstance(cstr, list):
                for c in cstr:
                    self._solver.assert_and_track(c, "asst_%s" % uuid.uuid4().hex[:8])
            else:
                self._solver.assert_and_track(cstr, "asst_%s" % uuid.uuid4().hex[:8])
        else:
            self._solver.add(cstr)

    def create_objective(self) -> bool:
        """create optimization objectives"""
        # in case of a single value to optimize
        if self.is_multi_objective_optimization_problem:
            # Replace objectives O_i, O_j, O_k with
            # O = WiOi+WjOj+WkOk etc.
            equivalent_single_objective = Int("EquivalentSingleObjective")
            weighted_objectives = []
            for obj in self.problem_context.objectives:
                variable_to_optimize = obj.target
                weight = obj.weight
                if isinstance(obj, MaximizeObjective):
                    weighted_objectives.append(-weight * variable_to_optimize)
                else:
                    weighted_objectives.append(weight * variable_to_optimize)
            self.add_constraint(equivalent_single_objective == Sum(weighted_objectives))
            # create an indicator
            equivalent_indicator = Indicator(
                "EquivalentIndicator", equivalent_single_objective
            )
            self.objective = MinimizeObjective(
                "EquivalentObjective", equivalent_indicator
            )
            self.add_constraint(equivalent_indicator.get_assertions())
        else:
            self.objective = self._problem.context.objectives[0]

    def process_work_amount(self) -> None:
        """for each task, compute the total work for all required resources"""
        for task in self.problem_context.tasks:
            if task.work_amount > 0.0:
                work_total_for_all_resources = []
                for required_resource in task.required_resources:
                    # work contribution for the resource
                    interv_low, interv_up = required_resource.busy_intervals[task]
                    work_contribution = required_resource.productivity * (
                        interv_up - interv_low
                    )
                    work_total_for_all_resources.append(work_contribution)
                self.add_constraint(
                    Sum(work_total_for_all_resources) >= task.work_amount
                )

    def check_sat(self):
        """check satisfiability. Returns resulta as True (sat) or False (unsat, unknown).
        The computation time.
        """
        init_time = time.perf_counter()
        sat_result = self._solver.check()
        check_sat_time = time.perf_counter() - init_time

        if sat_result == unsat:
            print(
                "\tNo solution can be found for problem %s.\n\tReason: Unsatisfiable problem: no solution exists"
                % self._problem.name
            )

        if sat_result == unknown:
            reason = self._solver.reason_unknown()
            print(
                "\tNo solution can be found for problem %s.\n\tReason: %s"
                % (self._problem.name, reason)
            )

        return sat_result, check_sat_time

    def build_solution(self, z3_sol):
        """create and return a SchedulingSolution instance"""
        solution = SchedulingSolution(self._problem)

        # set the horizon solution
        solution.horizon = z3_sol[self._problem.horizon].as_long()

        # process tasks
        for task in self._problem.context.tasks:
            # for each task, create a TaskSolution instance
            new_task_solution = TaskSolution(task.name)
            new_task_solution.type = type(task).__name__
            new_task_solution.start = z3_sol[task.start].as_long()
            new_task_solution.end = z3_sol[task.end].as_long()
            new_task_solution.duration = z3_sol[task.duration].as_long()
            new_task_solution.optional = task.optional

            # times, if ever delta_time and start_time are defined
            if self._problem.delta_time is not None:
                new_task_solution.duration_time = (
                    new_task_solution.duration * self._problem.delta_time
                )
                if self._problem.start_time is not None:
                    new_task_solution.start_time = (
                        self._problem.start_time
                        + new_task_solution.start * self._problem.delta_time
                    )
                    new_task_solution.end_time = (
                        new_task_solution.start_time + new_task_solution.duration_time
                    )
                else:
                    new_task_solution.start_time = (
                        new_task_solution.start * self._problem.delta_time
                    )
                    new_task_solution.end_time = (
                        new_task_solution.start_time + new_task_solution.duration_time
                    )

            if task.optional:
                # ugly hack, necessary because there's no as_bool()
                # method for Bool objects
                new_task_solution.scheduled = "%s" % z3_sol[task.scheduled] == "True"
            else:
                new_task_solution.scheduled = True

            # process resource assignments
            for req_res in task.required_resources:
                # by default, resource_should_be_assigned is set to True
                # if will be set to False if the resource is an alternative worker
                resource_is_assigned = True
                # among those workers, some of them
                # are busy "in the past", that is to say they
                # should not be assigned to the related task
                # for each interval
                lower_bound, _ = req_res.busy_intervals[task]
                if z3_sol[lower_bound].as_long() < 0:
                    # should not be scheduled
                    resource_is_assigned = False
                # add this resource to assigned resources, anytime
                if resource_is_assigned and (
                    req_res.name not in new_task_solution.assigned_resources
                ):
                    # if it is a cumulative resource, then we transform the resource name
                    resource_name = req_res.name.split("_CumulativeWorker_")[0]
                    if resource_name not in new_task_solution.assigned_resources:
                        new_task_solution.assigned_resources.append(resource_name)

            solution.add_task_solution(new_task_solution)

        # process resources
        for resource in self._problem.context.resources:
            # for each task, create a TaskSolution instance
            # for cumulative workers, we append the current work
            if "_CumulativeWorker_" in resource.name:
                cumulative_worker_name = resource.name.split("_CumulativeWorker_")[0]
                if cumulative_worker_name not in solution.resources:
                    new_resource_solution = ResourceSolution(cumulative_worker_name)
                else:
                    new_resource_solution = solution.resources[cumulative_worker_name]
            else:
                new_resource_solution = ResourceSolution(resource.name)
            new_resource_solution.type = type(resource).__name__
            # check for task processed by this resource
            for task in resource.busy_intervals.keys():
                task_name = task.name
                st_var, end_var = resource.busy_intervals[task]
                start = z3_sol[st_var].as_long()
                end = z3_sol[end_var].as_long()
                if (
                    start >= 0
                    and end >= 0
                    and (task_name, start, end) not in new_resource_solution.assignments
                ):
                    new_resource_solution.assignments.append((task_name, start, end))

            if "_CumulativeWorker_" in resource.name:
                cumulative_worker_name = resource.name.split("_CumulativeWorker_")[0]
                if cumulative_worker_name not in solution.resources:
                    solution.add_resource_solution(new_resource_solution)
            else:
                solution.add_resource_solution(new_resource_solution)

        # process indicators
        for indicator in self._problem.context.indicators:
            indicator_name = indicator.name
            indicator_value = z3_sol[indicator.indicator_variable].as_long()
            solution.add_indicator_solution(indicator_name, indicator_value)

        return solution

    def solve(self) -> Union[bool, SchedulingSolution]:
        """call the solver and returns the solution, if ever"""
        # for all cases
        if self.debug:
            self.print_assertions()

        if self.is_optimization_problem:
            if self.is_multi_objective_optimization_problem:
                print("\tObjectives:\n\t======")
                for obj in self._problem.context.objectives:
                    print("\t%s" % obj)
            # in this case, use the incremental solver
            if isinstance(self.objective, MinimizeObjective):
                dd = "min"
            elif isinstance(self.objective, MaximizeObjective):
                dd = "max"
            # print(dir(objective))
            # print(objective.target)
            solution = self.solve_optimize_incremental(self.objective.target, kind=dd)
            if not solution:
                return False
        else:
            # first check satisfiability
            sat_result, sat_computation_time = self.check_sat()

            print("Total computation time:\n=====================")
            print(
                "\t%s satisfiability checked in %.2fs"
                % (self._problem.name, sat_computation_time)
            )

            if sat_result == unsat:
                if self.debug:
                    unsat_core = self._solver.unsat_core()
                    print(
                        "\t%i unsatisfied assertion(s) (probable conflict):"
                        % len(unsat_core)
                    )
                    for c in unsat_core:
                        print("\t->%s" % c)
                return False

            if sat_result == unknown:
                return False

            # then get the solution
            solution = self._solver.model()

            # if self.objective:
            #     print('Optimization results:\n=====================')
            #     print('\t->Objective values:')
            #     for objective_name, objective_value in self.objectives:  # if ever no objectives, this line will do nothing
            #         print('\t\t->%s: %s' % (objective_name, objective_value.value()))

        self.current_solution = solution
        sol = self.build_solution(solution)

        if self.debug:
            self.print_statistics()
            self.print_solution()

        return sol

    def solve_optimize_incremental(
        self, variable: ArithRef, max_recursion_depth=None, kind="min"
    ) -> int:
        """target a min or max for a variable, without the Optimize solver.
        The loop continues ever and ever until the next value is more than 90%"""
        if kind not in ["min", "max"]:
            raise ValueError("choose either 'min' or 'max'")
        depth = 0
        solution = False
        total_time = 0
        current_variable_value = None
        print("Incremental optimizer:\n======================")
        three_last_times = []

        if self.objective.bounds is None:
            bound = None
        else:
            bound = (
                self.objective.bounds[0] if kind == "min" else self.objective.bounds[1]
            )

        while True:  # infinite loop, break if unsat of max_depth
            depth += 1
            if max_recursion_depth is not None:
                if depth > max_recursion_depth:
                    warnings.warn(
                        "maximum recursion depth exceeded. There might be a better solution."
                    )
                    break

            is_sat, sat_computation_time = self.check_sat()

            if is_sat == unsat and current_variable_value is not None:
                print(
                    "\tFound optimum %i. Stopping iteration." % current_variable_value
                )
                break
            elif is_sat == unsat and current_variable_value is None:
                print("\tNo solution found. Stopping iteration.")
                break
            elif is_sat == unknown:
                break
            # elif is_sat == unknown:
            #    return False
            # at this stage, is_sat should be sat
            solution = self._solver.model()

            current_variable_value = solution[variable].as_long()
            print(
                "\tFound better value:",
                current_variable_value,
                "elapsed time:%.3fs" % total_time,
            )
            total_time += sat_computation_time
            if total_time > self.max_time:
                warnings.warn("max time exceeded")
                break

            if bound is not None and current_variable_value == bound:
                print(
                    "\tFound optimum %i. Stopping iteration." % current_variable_value
                )
                break

            # prevent the solver to start a new round if we expect it to be
            # very long. The idea is the following: store the laste 3 computation
            # times, compute and extrapolation. Break the loop if ever the expected
            # time is too important.
            if len(three_last_times) < 3:
                three_last_times.append(total_time)
            else:
                three_last_times.pop(0)
                three_last_times.append(total_time)
                # Compute the expected value
                a, b, c = _calc_parabola_vertex([0, 1, 2], three_last_times)
                expected_next_time = a * 9 + 3 * b + c
                if expected_next_time > self.max_time:
                    warnings.warn("time may exceed max time. Stopping iteration.")
                    break
            self._solver.push()
            if kind == "min":
                self.add_constraint(variable < current_variable_value)
                print("\tChecking better value <%s" % current_variable_value)
            else:
                self.add_constraint(variable > current_variable_value)
                print("\tChecking better value >%s" % current_variable_value)

        print("\ttotal number of iterations: %i" % depth)
        if current_variable_value is not None:
            print("\tvalue: %i" % current_variable_value)
        print("\t%s satisfiability checked in %.2fs" % (self._problem.name, total_time))

        return solution

    def print_assertions(self):
        """A utility method to display solver assertions"""
        print("Assertions:\n===========")
        for assertion in self._solver.assertions():
            print("\t->", assertion)

    def print_statistics(self):
        """A utility method that displays solver statistics"""
        print("Solver satistics:")
        for key, value in self._solver.statistics():
            print("\t%s: %s" % (key, value))

    def print_solution(self):
        """A utility method that displays all internal variables for the current solution"""
        print("Solution:")
        for decl in self.current_solution.decls():
            var_name = decl.name()
            var_value = self.current_solution[decl]
            print("\t-> %s=%s" % (var_name, var_value))

    def find_another_solution(self, variable: ArithRef) -> bool:
        """let the solver find another solution for the variable"""
        if self.current_solution is None:
            warnings.warn("No current solution. First call the solve() method.")
            return False
        current_variable_value = self.current_solution[variable].as_long()
        self.add_constraint(variable != current_variable_value)
        return self.solve()

    def export_to_smt2(self, smt_filename):
        """export the model to a smt file to be processed by another SMT solver"""
        with open(smt_filename, "w") as outfile:
            outfile.write(self._solver.to_smt2())



# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import unittest

import processscheduler as ps


class ScheduleNTasksInTimeIntervals(unittest.TestCase):
    def test_single_interval_1(self) -> None:
        pb = ps.SchedulingProblem("ScheduleNTasksInTimeIntervals1", horizon=20)
        task_1 = ps.FixedDurationTask("task1", duration=3)
        task_2 = ps.FixedDurationTask("task2", duration=3)

        cstrt = ps.ScheduleNTasksInTimeIntervals(
            [task_1, task_2], nb_tasks_to_schedule=2, list_of_time_intervals=[[10, 13]]
        )
        pb.add_constraint(cstrt)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        # both tasks start and ends at the same time
        self.assertTrue(solution.tasks[task_1.name].start == 10)
        self.assertTrue(solution.tasks[task_1.name].end == 13)
        self.assertTrue(solution.tasks[task_2.name].start == 10)
        self.assertTrue(solution.tasks[task_2.name].end == 13)

    def test_single_interval_2(self) -> None:
        pb = ps.SchedulingProblem("ScheduleNTasksInTimeIntervals2", horizon=20)
        task_1 = ps.FixedDurationTask("task1", duration=3)
        task_2 = ps.FixedDurationTask("task2", duration=3)

        cstrt = ps.ScheduleNTasksInTimeIntervals(
            [task_1, task_2], nb_tasks_to_schedule=0, list_of_time_intervals=[[10, 20]]
        )
        pb.add_constraint(cstrt)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        # both tasks are necessarily scheduled before 10
        self.assertTrue(solution.tasks[task_1.name].end <= 10)
        self.assertTrue(solution.tasks[task_2.name].end <= 10)

    def test_single_interval_3(self) -> None:
        pb = ps.SchedulingProblem("ScheduleNTasksInTimeIntervals3", horizon=20)
        task_1 = ps.FixedDurationTask("task1", duration=3)
        task_2 = ps.FixedDurationTask("task2", duration=3)

        cstrt = ps.ScheduleNTasksInTimeIntervals(
            [task_1, task_2], nb_tasks_to_schedule=1, list_of_time_intervals=[[10, 20]]
        )
        pb.add_constraint(cstrt)
        # force task_1 to be shceduled after 10. So the only solution is that task 2 is scheduled
        # before 10
        cstrt1 = ps.TaskStartAfterLax(task_1, 10)
        pb.add_constraint(cstrt1)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        # both tasks are necessarily scheduled before 10
        self.assertTrue(solution.tasks[task_1.name].start >= 10)
        self.assertTrue(solution.tasks[task_2.name].end <= 10)

    def test_single_interval_no_solution(self) -> None:
        pb = ps.SchedulingProblem("ScheduleNTasksInTimeIntervalsNoSolution", horizon=20)
        task_1 = ps.FixedDurationTask("task1", duration=3)
        task_2 = ps.FixedDurationTask("task2", duration=3)

        cstrt = ps.ScheduleNTasksInTimeIntervals(
            [task_1, task_2],
            nb_tasks_to_schedule=3,  # impossible!!
            list_of_time_intervals=[[10, 20]],
        )
        pb.add_constraint(cstrt)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertFalse(solution)

    def test_single_interval_too_small(self) -> None:
        # no way to schedule tasks with duration 3 in a slot with range 2
        pb = ps.SchedulingProblem("ScheduleNTasksInTimeIntervalsTooSmall", horizon=20)
        task_1 = ps.FixedDurationTask("task1", duration=3)
        task_2 = ps.FixedDurationTask("task2", duration=3)

        cstrt = ps.ScheduleNTasksInTimeIntervals(
            [task_1, task_2], nb_tasks_to_schedule=2, list_of_time_intervals=[[10, 12]]
        )
        pb.add_constraint(cstrt)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertFalse(solution)

    def test_double_interval_1(self) -> None:
        pb = ps.SchedulingProblem(
            "ScheduleNTasksInTimeIntervalsDoubleInterval1", horizon=20
        )
        task_1 = ps.FixedDurationTask("task1", duration=3)

        cstrt = ps.ScheduleNTasksInTimeIntervals(
            [task_1], nb_tasks_to_schedule=1, list_of_time_intervals=[[5, 7], [15, 18]]
        )
        pb.add_constraint(cstrt)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertEqual(solution.tasks[task_1.name].start, 15)
        self.assertEqual(solution.tasks[task_1.name].end, 18)

    def test_double_interval_2(self) -> None:
        pb = ps.SchedulingProblem(
            "ScheduleNTasksInTimeIntervalsDoubleInterval2", horizon=20
        )
        task_1 = ps.FixedDurationTask("task1", duration=3)
        task_2 = ps.FixedDurationTask("task2", duration=3)

        cstrt = ps.ScheduleNTasksInTimeIntervals(
            [task_1, task_2],
            nb_tasks_to_schedule=2,
            list_of_time_intervals=[[5, 7], [15, 18]],
        )
        pb.add_constraint(cstrt)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)

    def test_triple_interval_1(self) -> None:
        pb = ps.SchedulingProblem(
            "ScheduleNTasksInTimeIntervalsTripleInterval1", horizon=20
        )
        task_1 = ps.FixedDurationTask("task1", duration=3)
        task_2 = ps.FixedDurationTask("task2", duration=3)

        cstrt = ps.ScheduleNTasksInTimeIntervals(
            [task_1, task_2],
            nb_tasks_to_schedule=2,
            list_of_time_intervals=[[5, 7], [11, 14], [15, 17]],
        )
        pb.add_constraint(cstrt)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertEqual(solution.tasks[task_1.name].start, 11)
        self.assertEqual(solution.tasks[task_1.name].end, 14)
        self.assertEqual(solution.tasks[task_2.name].start, 11)
        self.assertEqual(solution.tasks[task_2.name].end, 14)

    def test_triple_interval_that_overlap(self) -> None:
        pb = ps.SchedulingProblem(
            "ScheduleNTasksInTimeIntervalsTripleIntervalOverlap", horizon=20
        )
        task_1 = ps.FixedDurationTask("task1", duration=3)
        task_2 = ps.FixedDurationTask("task2", duration=3)

        cstrt = ps.ScheduleNTasksInTimeIntervals(
            [task_1, task_2],
            nb_tasks_to_schedule=2,
            list_of_time_intervals=[[5, 7], [6, 8], [15, 17]],
        )
        pb.add_constraint(cstrt)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertFalse(solution)

    def test_multi_constraintsp(self) -> None:
        # we want to schedule one task in slot [1,5] and one task into slot [7, 12]
        pb = ps.SchedulingProblem(
            "ScheduleNTasksInTimeIntervalsMultipleConstraints", horizon=20
        )
        task_1 = ps.FixedDurationTask("task1", duration=3)
        task_2 = ps.FixedDurationTask("task2", duration=3)

        cstrt1 = ps.ScheduleNTasksInTimeIntervals(
            [task_1, task_2], nb_tasks_to_schedule=1, list_of_time_intervals=[[1, 5]]
        )
        pb.add_constraint(cstrt1)

        cstrt2 = ps.ScheduleNTasksInTimeIntervals(
            [task_1, task_2], nb_tasks_to_schedule=1, list_of_time_intervals=[[7, 12]]
        )
        pb.add_constraint(cstrt2)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        # check the solution
        task1_start = solution.tasks[task_1.name].start
        task1_end = solution.tasks[task_1.name].end
        task2_start = solution.tasks[task_2.name].start
        task2_end = solution.tasks[task_2.name].end

        task1_in_interval_1 = task1_start >= 1 and task1_end <= 5
        task2_in_interval_1 = task2_start >= 1 and task2_end <= 5
        self.assertTrue(task1_in_interval_1 != task2_in_interval_1)  # xor

        task1_in_interval_2 = task1_start >= 7 and task1_end <= 12
        task2_in_interval_2 = task2_start >= 7 and task2_end <= 12
        self.assertTrue(task1_in_interval_2 != task2_in_interval_2)  # xor

        self.assertTrue(task1_in_interval_1 != task2_in_interval_1)  # xor
        self.assertTrue(task1_in_interval_2 != task2_in_interval_2)  # xor


if __name__ == "__main__":
    unittest.main()



"""Objective and indicator definition."""

# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Union, Tuple
from z3 import Int, BoolRef, ArithRef

from processscheduler.base import _NamedUIDObject
import processscheduler.context as ps_context


class Indicator(_NamedUIDObject):
    """an performance indicator, can be evaluated after the solver has finished solving,
    or being optimized (Max or Min) *before* calling the solver."""

    def __init__(
        self,
        name: str,
        expression: Union[BoolRef, ArithRef],
        bounds: Optional[Tuple[Optional[int]]] = None,
    ) -> None:
        super().__init__(name)
        # scheduled start, end and duration set to 0 by default
        # be set after the solver is called
        if not isinstance(expression, (BoolRef, ArithRef)):
            raise TypeError(
                "the indicator expression must be either a BoolRef or ArithRef."
            )
        self.name = name
        self.indicator_variable = Int("Indicator_%s" % name)
        # by default the scheduled value is set to None
        # set by the solver
        self.scheduled_value = None

        # -- Bound --
        # None if not bounded
        # (lower_bound, None), (None, upper_bound) if only one-side bounded
        # (lower_bound, upper_bound) if full bounded
        self.bounds = bounds

        self.add_assertion(self.indicator_variable == expression)

        ps_context.main_context.add_indicator(self)


class Objective(_NamedUIDObject):
    def __init__(self, name: str, target: Union[ArithRef, Indicator]) -> None:
        super().__init__(name)
        if not isinstance(target, (ArithRef, Indicator)):
            raise TypeError(
                "the indicator expression must be either a BoolRef, ArithRef or Indicator instance."
            )
        if isinstance(target, Indicator):
            self.target = target.indicator_variable
            self.bounds = target.bounds
        else:
            self.target = target
            self.bounds = None
        ps_context.main_context.add_objective(self)


class MaximizeObjective(Objective):
    def __init__(
        self, name: str, target: Union[ArithRef, Indicator], weight: Optional[int] = 1
    ) -> None:
        super().__init__(name, target)
        self.weight = weight


class MinimizeObjective(Objective):
    def __init__(
        self, name: str, target: Union[ArithRef, Indicator], weight: Optional[int] = 1
    ) -> None:
        super().__init__(name, target)
        self.weight = weight
